"""
Created on 10 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
    "format": "application/json"
}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class TopicInfo(JSONable):
    """
    classdocs
   """

    FORMAT_JSON = "application/json"
    FORMAT_TEXT = "text"


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        topic_format = jdict.get('format')
        topic_type = jdict.get('type')
        schema = jdict.get('schema')
        graph_path = jdict.get('graph-path')

        return TopicInfo(topic_format, topic_type, schema, graph_path)


    # ----------------------------------------------------------------------------------------------------------------
    # for the v2 API, schema_id goes in Topic...

    def __init__(self, topic_format, topic_type=None, schema=None, graph_path=None):
        """
        Constructor
        """
        self.__format = topic_format        # string
        self.__type = topic_type            # string
        self.__schema = schema              # string
        self.__graph_path = graph_path      # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.format:
            jdict['format'] = self.format

        if self.type:
            jdict['type'] = self.type

        if self.schema:
            jdict['schema'] = self.schema

        if self.graph_path:
            jdict['graph-path'] = self.graph_path

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def schema_id(self):
        if self.__schema is None:
            return None

        return self.__schema.id


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def format(self):
        return self.__format


    @property
    def type(self):
        return self.__type


    @property
    def schema(self):
        return self.__schema


    @property
    def graph_path(self):
        return self.__graph_path


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicInfo:{format:%s, type:%s, schema:%s, graph_path:%s}" % \
                    (self.format, self.type, self.schema, self.graph_path)



"""
Created on 20 Jun 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense A4 IRC-AT non-dispersive infra-red detector
"""

from collections import OrderedDict
from numbers import Number

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class NDIRDatum(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        temp = jdict.get('tmp')
        cnc = jdict.get('cnc-raw')
        cnc_igl = jdict.get('cnc')

        return NDIRDatum(temp, cnc, cnc_igl)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, temp, cnc, cnc_igl):
        """
        Constructor
        """
        self.__temp = Datum.float(temp, 1)              # temperature                               °C
        self.__cnc = Datum.float(cnc, 1)                # concentration                             ppm
        self.__cnc_igl = Datum.float(cnc_igl, 1)        # concentration (ideal gas law corrected)   ppm


    def __eq__(self, other):
        try:
            return self.temp == other.temp and self.cnc == other.cnc and self.cnc_igl == other.cnc_igl

        except AttributeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------
    # Support for averaging...

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(other)

        temp = self.temp + other.temp
        cnc = self.cnc + other.cnc
        cnc_igl = self.cnc_igl + other.cnc_igl

        return NDIRDatum(temp, cnc, cnc_igl)


    def __truediv__(self, other):
        if not isinstance(other, Number):
            raise TypeError(other)

        temp = self.temp / other
        cnc = None if self.cnc is None else self.cnc / other
        cnc_igl = None if self.cnc_igl is None else self.cnc_igl / other

        return NDIRDatum(temp, cnc, cnc_igl)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tmp'] = self.temp
        jdict['cnc-raw'] = self.cnc
        jdict['cnc'] = self.cnc_igl

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def temp(self):
        return self.__temp


    @property
    def cnc(self):
        return self.__cnc


    @property
    def co2(self):              # a synonym of cnc for compatibility with GasesSample and SCD30Datum
        return self.cnc_igl


    @property
    def cnc_igl(self):
        return self.__cnc_igl


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "NDIRDatum:{temp:%s, cnc:%s, cnc_igl:%s}" % (self.temp, self.cnc, self.cnc_igl)



"""
Created on 25 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Equivalent to cURLs:
curl "https://aws.southcoastscience.com/device-topics?topic=south-coast-science-dev/alphasense/loc/303/gases"
curl "https://aws.southcoastscience.com/device-topics?device=scs-bgx-303"
"""

from scs_core.aws.client.rest_client import RESTClient
from scs_core.aws.data.byline import Byline, DeviceBylineGroup, TopicBylineGroup


# --------------------------------------------------------------------------------------------------------------------

class BylineManager(object):
    """
    classdocs
    """

    __DEVICE =      'device'
    __TOPIC =       'topic'


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find_latest_byline_for_topic(self, topic):
        request_path = '/device-topics'

        params = {self.__TOPIC: topic}

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path, params=params)

            # bylines...
            if jdict is None:
                return None

            latest_byline = None

            for item in jdict:
                byline = Byline.construct_from_jdict(item)

                if latest_byline is None or latest_byline.rec < byline.rec:
                    latest_byline = byline

            return latest_byline

        finally:
            self.__rest_client.close()



    def find_bylines(self, excluded=None):
        request_path = '/device-topics'

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path)

            # bylines...
            return TopicBylineGroup.construct_from_jdict(jdict, excluded=excluded, skeleton=True)

        finally:
            self.__rest_client.close()


    def find_bylines_for_topic(self, topic, excluded=None):
        request_path = '/device-topics'

        params = {self.__TOPIC: topic}

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path, params=params)

            # bylines...
            return TopicBylineGroup.construct_from_jdict(jdict, excluded=excluded, skeleton=True)

        finally:
            self.__rest_client.close()


    def find_bylines_for_device(self, device, excluded=None):
        request_path = '/device-topics'

        params = {self.__DEVICE: device}

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path, params=params)

            # bylines...
            return DeviceBylineGroup.construct_from_jdict(jdict, excluded=excluded, skeleton=True)

        finally:
            self.__rest_client.close()


    def find_byline_for_device_topic(self, device, topic):
        request_path = '/device-topics'

        params = {self.__DEVICE: device}

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path, params=params)

            # bylines...
            if jdict is None:
                return None

            for item in jdict:
                byline = Byline.construct_from_jdict(item)

                if byline.topic == topic:
                    return byline

            return None

        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "BylineManager:{rest_client:%s}" % self.__rest_client



"""
Created on 16 Sep 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Timeout context manager. Usage:

timeout = Timeout(10)

try:
    with timeout:
        my_func()

except TimeoutError:
    print("function timed out!")

https://www.jujens.eu/posts/en/2018/Jun/02/python-timeout-function/
"""

import signal


# --------------------------------------------------------------------------------------------------------------------

class Timeout(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, seconds):
        """
        Constructor
        """
        self.__seconds = seconds                                    # number


    def __enter__(self):
        # handler...
        def raise_timeout(_signum, _frame):
            signal.signal(signal.SIGALRM, signal.SIG_IGN)
            raise TimeoutError

        # signal...
        signal.signal(signal.SIGALRM, raise_timeout)
        signal.alarm(int(round(self.__seconds)))


    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        signal.signal(signal.SIGALRM, signal.SIG_IGN)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Timeout:{seconds:%s}" % self.__seconds



"""
Created on 10 Dec 2020
Author: Jade Page (jade.page@southcoastscience.com)

Like lambda_message_manager, but speaks directly to the lambda
"""

# --------------------------------------------------------------------------------------------------------------------
import json
from urllib.parse import urlparse, parse_qs

from scs_core.aws.manager.lambda_message_manager import MessageResponse


class MessageManager(object):
    """
    classdocs
    """
    __TOPIC = 'topic'
    __START = 'startTime'
    __END = 'endTime'
    __REC_ONLY = 'rec_only'

    def __init__(self, lambda_client):
        """
        Constructor
        """
        self.__lambda_client = lambda_client

    def find_for_topic(self, topic, start_date, end_date):

        event = {'topic': topic,
                 'startTime': start_date.utc().as_iso8601(True),
                 'endTime': end_date.utc().as_iso8601(True)}

        while True:
            # request...
            payload = json.dumps(event).encode('utf-8')
            response = self.__lambda_client.invoke(
                FunctionName="arn:aws:lambda:us-west-2:696437392763:function:topicMessages",
                InvocationType='RequestResponse',
                Payload=payload
            )

            res = response['Payload'].read()

            j_res = json.loads(res.decode('utf-8'))
            body = json.loads(j_res.get("body"))

            # messages...
            block = MessageResponse.construct_from_jdict(body)

            for item in block.items:
                yield item

            if block.next_url is None:
                break

            next_url = urlparse(block.next_url)
            next_params = parse_qs(next_url.query)

            event = {'topic': topic,
                     'startTime': next_params[self.__START][0],
                     'endTime': end_date.utc().as_iso8601(True)}





"""
Created on 23 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"rec": "2021-10-06T11:27:48Z", "tag": "scs-be2-3", "ver": 1.00, "val": {"hostname": "scs-bbe-003",
"packs": {"scs_comms": {"repo": "scs_comms_ge910", "version": null}, "scs_core": {"repo": "scs_core",
"version": "1.0.32"}, "scs_dev": {"repo": "scs_dev", "version": "1.0.11"}, "scs_dfe": {"repo": "scs_dfe_eng",
"version": "1.0.10"}, "scs_exegesis": {"repo": "scs_exegesis", "version": null},
"scs_greengrass": {"repo": "scs_greengrass", "version": "1.0.3"}, "scs_host": {"repo": "scs_host_bbe_southern",
"version": "1.0.6"}, "scs_inference": {"repo": "scs_inference", "version": null}, "scs_mfr": {"repo": "scs_mfr",
"version": "1.0.14"}, "scs_ndir": {"repo": "scs_ndir", "version": null}, "scs_osio": {"repo": "scs_osio",
"version": null}, "scs_psu": {"repo": "scs_psu", "version": "1.0.10"}},
"afe-baseline": {"sn1": {"calibrated-on": "2021-08-22T12:46:44Z", "offset": 13, "env": {"hmd": 62.6, "tmp": 24.5,
"pA": 102.1}}, "sn2": {"calibrated-on": "2021-08-22T12:46:38Z", "offset": -7, "env": {"hmd": 62.7, "tmp": 24.5,
"pA": 102.1}}, "sn3": {"calibrated-on": "2021-08-22T12:46:46Z", "offset": 24, "env": {"hmd": 62.6, "tmp": 24.5,
"pA": 102.1}}, "sn4": {"calibrated-on": "2021-08-22T12:46:41Z", "offset": 9, "env": {"hmd": 62.6, "tmp": 24.5,
"pA": 102.1}}}, "afe-id": {"serial_number": "27-000001", "type": "810-0023-02", "calibrated_on": "2016-11-01",
"sn1": {"serial_number": "212060308", "sensor_type": "NO2A43F"}, "sn2": {"serial_number": "132950202",
"sensor_type": "CO A4"}, "sn3": {"serial_number": "134060009", "sensor_type": "SO2A4"},
"sn4": {"serial_number": "133910023", "sensor_type": "H2SA4"}},
"aws-api-auth": {"endpoint": "aws.southcoastscience.com", "api-key": "a04c-62d684d64a1f"},
"aws-client-auth": {"endpoint": "asrfh6e5j5ecz.iot.us-west-2.amazonaws.com", "client-id": "scs-bbe-003",
"cert-id": "cd505d98bf"}, "aws-group-config": {"group-name": "scs-bbe-003-group",
"time-initiated": "2021-01-29T11:52:25Z", "unix-group": 987, "ml": true},
"aws-project": {"location-path": "south-coast-science-dev/development/loc/1",
"device-path": "south-coast-science-dev/development/device"},
"csv-logger-conf": {"root-path": "/srv/removable_data_storage", "delete-oldest": true, "write-interval": 0},
"display-conf": null, "gas-baseline": {"NO2": {"calibrated-on": "2021-08-22T12:40:13Z", "offset": 3,
"env": {"hmd": 62.5, "tmp": 24.3, "pA": 102.0}}}, "gas-model-conf": {"uds-path": "pipes/lambda-gas-model.uds",
"model-interface": "vB"}, "gps-conf": {"model": "SAM8Q", "sample-interval": 10, "tally": 60,
"report-file": "/tmp/southcoastscience/gps_report.json", "debug": false}, "interface-conf": {"model": "DFE"},
"greengrass-identity": null, "mpl115a2-calib": {"calibrated-on": "2020-11-15T11:29:23Z", "c25": 510},
"mqtt-conf": {"inhibit-publishing": true, "report-file": null, "debug": false}, "ndir-conf": {"model": "t1f1",
"tally": 1, "raw": false}, "opc-conf": {"model": "N3", "sample-period": 10, "restart-on-zeroes": true,
"power-saving": false}, "opc-version": null, "pmx-model-conf": {"uds-path": "pipes/lambda-pmx-model.uds",
"model-interface": "s1"}, "pressure-conf": {"model": "ICP", "altitude": 101}, "psu-conf": {"model": "OsloV1",
"batt-model": null, "ignore-threshold": false, "reporting-interval": 20,
"report-file": "/tmp/southcoastscience/psu_status_report.json"}, "psu-version": null,
"pt1000-calib": {"calibrated-on": "2017-08-15T11:21:45Z", "v20": 0.320208},
"scd30-baseline": {"CO2": {"calibrated-on": "2021-08-22T12:40:15Z", "offset": 5, "env": {"hmd": 62.5, "tmp": 24.3,
"pA": 102.0}}}, "scd30-conf": null, "schedule": {"scs-climate": {"interval": 60.0, "tally": 1},
"scs-gases": {"interval": 10.0, "tally": 1}, "scs-status": {"interval": 60.0, "tally": 1}},
"shared-secret": {"key": "pYL7B1JcgJ2gy6MP"}, "sht-conf": {"int": "0x45", "ext": "0x45"},
"networks": {"eth0": {"kind": "ethernet", "state": "connected", "connection": "Ethernet eth0"},
"usb0": {"kind": "ethernet", "state": "unavailable", "connection": null}}, "modem": null, "sim": null,
"system-id": {"set-on": "2021-09-12T12:04:25Z", "vendor-id": "SCS", "model-id": "BE2", "model": "Alpha BB Eng",
"config": "V2", "system-sn": 3}, "timezone-conf": {"set-on": "2021-01-31T11:26:14Z", "name": "Europe/London"}}}
"""

import json

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str

from scs_core.estate.configuration import Configuration

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationSample(Sample):
    """
    classdocs
    """

    VERSION = 1.0

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # Sample...
        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        version = jdict.get('ver', cls.DEFAULT_VERSION)

        try:
            val_jdict = json.loads(jdict.get('val'))
        except TypeError:
            val_jdict = jdict.get('val')

        configuration = Configuration.construct_from_jdict(val_jdict)

        return cls(tag, rec, configuration, version=version)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, configuration, version=None):
        """
        Constructor
        """
        if version is None:
            version = self.VERSION

        super().__init__(tag, rec, version)

        self.__configuration = configuration                # Configuration


    def __eq__(self, other):
        try:
            return self.tag == other.tag and self.configuration == other.configuration

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        if self.tag < other.tag:
            return True

        if self.tag > other.tag:
            return False

        if self.rec < other.rec:
            return True

        return False


    def diff(self, other):
        return ConfigurationSample(self.tag, self.rec, self.configuration.diff(other.configuration))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        return self.configuration.as_json()


    @property
    def configuration(self):
        return self.__configuration


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationSample:{tag:%s, rec:%s, version:%s, configuration:%s}" % \
               (self.tag, self.rec, self.version, self.configuration)


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationSampleHistory(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, latest_only=False, items=None):
        """
        Constructor
        """
        self.__latest_only = latest_only                            # bool
        self.__items = {} if items is None else items               # dict of tag: list of ConfigurationSample


    # ----------------------------------------------------------------------------------------------------------------

    def insert(self, sample: ConfigurationSample):
        if sample.tag not in self.__items:
            self.__items[sample.tag] = []
            self.__items[sample.tag].append(sample)
            return

        if not self.__items[sample.tag] or not self.__latest_only:
            if sample not in self.__items[sample.tag]:            # we might not be reading all the fields in the DB!
                self.__items[sample.tag].append(sample)
            return

        for item in self.__items[sample.tag]:
            if sample.rec > item.rec:
                self.__items[sample.tag] = (sample, )
                return


    # ----------------------------------------------------------------------------------------------------------------

    def diffs(self):
        diff_history = ConfigurationSampleHistory()

        for tag in self.tags():
            prev = None
            for item in self.items_for_tag(tag):
                diff_history.insert(item if prev is None else item.diff(prev))
                prev = item

        return diff_history


    def tags(self):
        return set(self.__items.keys())


    def items_for_tag(self, tag):
        if tag not in self.__items:
            return None

        return self.__items[tag]


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        items = []

        for tag in self.tags():
            items += self.items_for_tag(tag)

        return items


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationSampleHistory:{latest_only:%s, items:%s}" % \
               (self.__latest_only, Str.collection(self.__items))


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationReport(ConfigurationSample):
    """
    classdocs
    """

    @classmethod
    def construct(cls, sample: ConfigurationSample, report: LocalizedDatetime):
        return cls(sample.tag, sample.rec, sample.configuration, report)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, configuration, report):
        """
        Constructor
        """
        super().__init__(tag, rec, configuration)

        self.__report = report                          # LocalizedDatetime


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['rec'] = {'report': self.report.as_iso8601(), 'update': self.rec.as_iso8601()}
        jdict['ver'] = round(self.version, 1)

        jdict['val'] = self.values

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def report(self):
        return self.__report


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationReport:{tag:%s, rec:%s, configuration:%s, report:%s}" % \
               (self.tag, self.rec, self.configuration, self.report)





"""
Created on 30 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
  "id": "southcoastscience-dev",
  "name": "South Coast Science - Dev",
  "month": "2016-11",
  "gravatar-hash": "07f512e9fe64863039df0c0f1834cc25",
  "topics": [
    {
      "topic": "/orgs/south-coast-science-dev/user/device/alpha-pi-eng-000100/status",
      "name": "Device status",
      "description": "lat (deg), lng (deg) GPS qual, DFE temp (Centigrade), host temp (Centigrade), errors",
      "public": true,
      "topic-info": {
        "format": "application/json"
      },
      "unit": null,
      "stats": {
        "period": 0,
        "last-reading": "55 days ago",
        "contributors": [
          {
            "name": "South Coast Science - Dev",
            "id": "southcoastscience-dev",
            "gravatar-hash": "07f512e9fe64863039df0c0f1834cc25"
          }
        ],
        "last-location": {
          "lat": 50.82313,
          "lon": -0.122922
        }
      },
      "owner-gravatar-hash": "d92c0a697bf3658745447e1a9467b506"
    }
    ]}
"""

from scs_core.data.str import Str

from scs_core.osio.data.user import User
from scs_core.osio.data.user_topic import UserTopic


# --------------------------------------------------------------------------------------------------------------------

class UserMetadata(User):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # User...
        id = jdict.get('id')
        name = jdict.get('name')

        start = jdict.get('month')

        # UserMetadata...
        gravatar_hash = jdict.get('gravatar-hash')

        topics = [UserTopic.construct_from_jdict(topic_jdict) for topic_jdict in jdict.get('topics')]

        return UserMetadata(id, name, None, None, start, gravatar_hash, topics)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, name, email, password, start, gravatar_hash, topics):
        """
        Constructor
        """

        # User...
        User.__init__(self, id, name, email, password, start)

        # UserMetadata...
        self.__gravatar_hash = gravatar_hash            # string
        self.__topics = topics                          # list of UserTopic


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = User.as_json(self)

        jdict['gravatar-hash'] = self.gravatar_hash
        jdict['topics'] = self.topics

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def gravatar_hash(self):
        return self.__gravatar_hash


    @property
    def topics(self):
        return self.__topics


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "UserMetadata:{id:%s, name:%s, email:%s, password:%s, start:%s, gravatar_hash:%s, " \
               "topics:%s}" % \
               (self.id, self.name, self.email, self.password, self.start, self.gravatar_hash,
                Str.collection(self.topics))



"""
Created on 21 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from abc import ABC
from collections import OrderedDict


# --------------------------------------------------------------------------------------------------------------------

class SFTPClientConf(ABC):
    """
    classdocs
    """

    DEFAULT_PORT = 22

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        host = jdict.get('host')
        port = jdict.get('port')

        username = jdict.get('username')
        password = jdict.get('password')

        remote_path = jdict.get('remote-path')

        return cls(host, port, username, password, remote_path)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host, port, username, password, remote_path):
        """
        Constructor
        """
        self.__host = str(host)                             # string
        self.__port = int(port)                             # int

        self.__username = str(username)                     # string
        self.__password = str(password)                     # string

        self.__remote_path = remote_path                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['host'] = self.host
        jdict['port'] = self.port

        jdict['username'] = self.username
        jdict['password'] = self.password

        jdict['remote-path'] = self.remote_path

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def host(self):
        return self.__host


    @property
    def port(self):
        return self.__port


    @property
    def username(self):
        return self.__username


    @property
    def password(self):
        return self.__password


    @property
    def remote_path(self):
        return self.__remote_path


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{host:%s, port:%s, username:%s, password:%s, remote_path:%s}" %  \
            (self.host, self.port, self.username, self.password, self.remote_path)



"""
Created on 24 Oct 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict
from numbers import Number

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class NDIRVoltages(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        ref = jdict.get('ref')
        act = jdict.get('act')
        therm = jdict.get('therm')

        return NDIRVoltages(ref, act, therm)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, ref, act, therm):
        """
        Constructor
        """
        self.__ref = Datum.float(ref, 6)                # pile_ref_amplitude        Volts
        self.__act = Datum.float(act, 6)                # pile_act_amplitude        Volts
        self.__therm = Datum.float(therm, 6)            # thermistor_average        Volts


    def __eq__(self, other):
        try:
            return self.ref == other.ref and self.act == other.act and self.therm == other.therm

        except AttributeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------
    # Support for averaging...

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(other)

        ref = self.ref + other.ref
        act = self.act + other.act
        therm = self.therm + other.therm

        return NDIRVoltages(ref, act, therm)


    def __truediv__(self, other):
        if not isinstance(other, Number):
            raise TypeError(other)

        ref = self.ref / other
        act = None if self.act is None else self.act / other
        therm = None if self.therm is None else self.therm / other

        return NDIRVoltages(ref, act, therm)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['ref'] = self.ref
        jdict['act'] = self.act
        jdict['therm'] = self.therm

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def ref(self):
        return self.__ref


    @property
    def act(self):
        return self.__act


    @property
    def therm(self):
        return self.__therm


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "NDIRVoltages:{ref:%s, act:%s, therm:%s}" % (self.ref, self.act, self.therm)



"""
Created on 17 May 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Note: time shall always be represented as UTC, then localized on conversion.
"""

from datetime import datetime
from datetime import timedelta
from datetime import timezone

import re

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class RTCDatetime(JSONable):
    """
    classdocs
    """

    CENTURY =               2000


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr):
        if not jstr:
            return None

        match = re.match(r'(\d{2})-(\d{2})-(\d{2}) \((\d)\) (\d{2}):(\d{2}):(\d{2})', jstr)

        if match is None:
            return None

        fields = match.groups()

        # fields...
        year = int(fields[0])
        month = int(fields[1])
        day = int(fields[2])

        weekday = int(fields[3])

        hour = int(fields[4])
        minute = int(fields[5])
        second = int(fields[6])

        return cls(year, month, day, weekday, hour, minute, second)


    @classmethod
    def construct_from_localized_datetime(cls, localized_datetime):
        # RTC zone...
        zone_offset = timedelta(hours=0, minutes=0)
        utc_zone = timezone(zone_offset)

        # localized...
        utc = localized_datetime.localize(utc_zone)
        rtc = utc.datetime

        # fields...
        year = rtc.year % 100
        month = rtc.month
        day = rtc.day

        weekday = rtc.isoweekday()

        hour = rtc.hour
        minute = rtc.minute
        second = rtc.second

        return cls(year, month, day, weekday, hour, minute, second)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, year, month, day, weekday, hour, minute, second):
        """
        Constructor
        """
        self.__year = year              # int       0 - 99
        self.__month = month            # int       1 - 12
        self.__day = day                # int       1 - 31

        self.__weekday = weekday        # int       1 - 7   ISO 8601: Monday is 1

        self.__hour = hour              # int       0 - 23
        self.__minute = minute          # int       0 - 59
        self.__second = second          # int       0 - 59


    # ----------------------------------------------------------------------------------------------------------------

    def as_localized_datetime(self, local_zone):        # may be pytz timezone or datetime timezone
        # RTC zone...
        zone_offset = timedelta(hours=0, minutes=0)
        utc_zone = timezone(zone_offset)

        # localized...
        year = RTCDatetime.CENTURY + self.year

        rtc = datetime(year, month=self.month, day=self.day, hour=self.hour, minute=self.minute, second=self.second,
                       tzinfo=utc_zone)
        utc = LocalizedDatetime(rtc)

        # ...to host zone...
        localized = utc.localize(local_zone)

        return localized


    def as_json(self):
        return "%02d-%02d-%02d (%d) %02d:%02d:%02d" % \
               (self.year, self.month, self.day, self.weekday, self.hour, self.minute, self.second)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def year(self):
        return self.__year


    @property
    def month(self):
        return self.__month


    @property
    def day(self):
        return self.__day


    @property
    def weekday(self):
        return self.__weekday


    @property
    def hour(self):
        return self.__hour


    @property
    def minute(self):
        return self.__minute


    @property
    def second(self):
        return self.__second


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "RTCDatetime:{year:%s, month:%s, day:%s, weekday:%s, hour:%s, minute:%s, second:%s}" % \
               (self.year, self.month, self.day, self.weekday, self.hour, self.minute, self.second)



"""
Created on 30 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from abc import ABC

from scs_core.data.json import JSONable
from scs_core.gas.sensor import Sensor


# --------------------------------------------------------------------------------------------------------------------

class SensorCalib(ABC, JSONable):
    """
    classdocs
    """

    ALPHASENSE_HOST =       "www.alphasense-technology.co.uk"
    ALPHASENSE_PATH =       "/api/v1/sensors/"
    ALPHASENSE_HEADER =     {"Accept": "application/json"}


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        from scs_core.gas.a4.a4_calib import A4Calib            # late import
        from scs_core.gas.pid.pid_calib import PIDCalib         # late import

        sensor_type = jdict.get('sensor_type', 'NOGA4')

        if 'A4' in sensor_type or sensor_type.startswith('SN'):
            return A4Calib.construct_from_jdict(jdict)

        elif sensor_type.startswith('PID'):
            return PIDCalib.construct_from_jdict(jdict)

        raise ValueError(sensor_type)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def reports_no2_cross_sensitivity(cls):                     # the default - override as necessary
        return False


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, sensor_type):
        """
        Constructor
        """
        self.__serial_number = serial_number                    # int
        self.__sensor_type = sensor_type                        # string


    def __eq__(self, other):
        try:
            return self.serial_number == other.serial_number and self.sensor_type == other.sensor_type

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def sensor(self, baseline):         # TODO: do this on the subclass?
        sensor = Sensor.find(self.__serial_number)

        sensor.calib = self
        sensor.baseline = baseline

        # TODO: set the calibrator - if it is an electrochem! (vCal baseline belongs to the inference system?)

        return sensor


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def serial_number(self):
        return self.__serial_number


    @serial_number.setter
    def serial_number(self, serial_number):
        self.__serial_number = serial_number


    @property
    def sensor_type(self):
        return self.__sensor_type


    @sensor_type.setter
    def sensor_type(self, sensor_type):
        self.__sensor_type = sensor_type




"""
Created on 22 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class Sample(JSONable):
    """
    classdocs
    """

    DEFAULT_VERSION = 1.0

    EXEGESIS_TAG =              "exg"
    INCLUDE_MILLIS =            False

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        version = jdict.get('ver', cls.DEFAULT_VERSION)

        src = jdict.get('src')
        values = jdict.get('val')
        exegeses = jdict.get('exg')

        return cls(tag, rec, version, src=src, values=values, exegeses=exegeses)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, version, src=None, values=None, exegeses=None):
        """
        Constructor
        """
        self.__tag = tag                                # string
        self.__rec = rec                                # LocalizedDatetime

        self.__version = float(version)                 # float
        self.__src = src                                # string

        self.__values = values                          # OrderedDict
        self.__exegeses = exegeses                      # OrderedDict


    def __eq__(self, other):
        try:
            return self.tag == other.tag and self.rec == other.rec and \
                   self.version == other.version and self.src == other.src and \
                   self.values == other.values and self.exegeses == other.exegeses

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['rec'] = None if self.rec is None else self.rec.as_iso8601(include_millis=self.INCLUDE_MILLIS)

        if self.tag is not None:
            jdict['tag'] = self.tag

        jdict['ver'] = round(self.version, 1)

        if self.src is not None:
            jdict['src'] = self.src

        jdict['val'] = self.values

        if self.exegeses:
            jdict[self.EXEGESIS_TAG] = self.exegeses

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def rec(self):
        return self.__rec


    @rec.setter
    def rec(self, rec):
        self.__rec = rec


    @property
    def version(self):
        return self.__version


    @version.setter
    def version(self, version):
        self.__version = version


    @property
    def src(self):
        return self.__src


    @property
    def values(self):
        return self.__values


    @property
    def exegeses(self):
        return self.__exegeses


    @exegeses.setter
    def exegeses(self, exegeses):
        self.__exegeses = exegeses


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        values = Str.collection(self.values)
        exegeses = Str.collection(self.exegeses)

        return self.__class__.__name__ + ":{tag:%s, rec:%s, version:%s, src:%s, values:%s, exegeses:%s}" % \
            (self.tag, self.rec, self.version, self.src, values, exegeses)



"""
Created on 11 May 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://pymotw.com/2/subprocess/
"""

import os

from collections import OrderedDict
from subprocess import Popen, PIPE, TimeoutExpired

from scs_core.data.json import JSONable, JSONify


# --------------------------------------------------------------------------------------------------------------------

class Command(JSONable):
    """
    classdocs
    """

    __LIST_CMD = '?'

    __PROHIBITED_TOKENS = ('-i', '--interactive', '<', '>', ';', '|')

    __DEFAULT_TIMEOUT = 30.0            # seconds


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        cmd = jdict.get('cmd')
        params = jdict.get('params')

        stdout = jdict.get('stdout')
        stderr = jdict.get('stderr')
        return_code = jdict.get('ret')

        datum = cls(cmd, params, stdout=stdout, stderr=stderr, return_code=return_code)

        return datum


    @classmethod
    def construct_from_tokens(cls, tokens):
        if not tokens:
            return cls(None, [])

        cmd = tokens[0]
        params = tokens[1:]

        return cls(cmd, params)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, cmd, params, stdout=None, stderr=None, return_code=None):
        """
        Constructor
        """
        self.__cmd = cmd                        # string
        self.__params = params                  # array

        self.__stdout = stdout                  # array of string
        self.__stderr = stderr                  # array of string
        self.__return_code = return_code        # int


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self, host):
        try:
            host.command_path()
        except NotImplementedError:
            return False

        if set(self.params).intersection(set(Command.__PROHIBITED_TOKENS)):
            return False

        if self.cmd == Command.__LIST_CMD:
            return True

        if self.cmd not in os.listdir(host.command_path()):
            return False

        return True


    def execute(self, host, timeout):
        if not self.cmd:
            return None

        try:
            if self.cmd == Command.__LIST_CMD:
                result = self.__execute('ls', host, timeout)

                self.__stdout = [JSONify.dumps(self.__stdout)]

            else:
                statement = ['./' + self.cmd]
                statement.extend(self.params)

                result = self.__execute(statement, host, timeout)

        except OSError as ex:
            return self.error(repr(ex))

        return result


    def error(self, message):
        self.__stdout = []
        self.__stderr = [message]
        self.__return_code = 1


    # ----------------------------------------------------------------------------------------------------------------

    def __execute(self, statement, host, timeout):
        p = Popen(statement, cwd=host.command_path(), stdout=PIPE, stderr=PIPE)

        try:
            stdout_bytes, stderr_bytes = p.communicate(timeout=timeout)

            self.__stdout = stdout_bytes.decode().strip().splitlines()
            self.__stderr = stderr_bytes.decode().strip().splitlines()
            self.__return_code = p.returncode

            return self.__return_code == 0

        except TimeoutExpired as ex:
            self.error(repr(ex))
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['cmd'] = self.cmd
        jdict['params'] = self.params

        jdict['stdout'] = self.stdout
        jdict['stderr'] = self.stderr
        jdict['ret'] = self.return_code

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def cmd(self):
        return self.__cmd


    @property
    def params(self):
        return self.__params


    @property
    def stdout(self):
        return self.__stdout


    @property
    def stderr(self):
        return self.__stderr


    @property
    def return_code(self):
        return self.__return_code


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Command:{cmd:%s, params:%s, stdout:%s, stderr:%s, return_code:%s}" % \
               (self.cmd, self.params, self.stdout, self.stderr, self.return_code)



"""
Created on 13 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

the I2C addresses of the internal (in A4 pot) and external (exposed to air) SHTs

example JSON:
{"int": "0x44", "ext": "0x45"}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class SHTConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "sht_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __addr_str(cls, addr):
        if addr is None:
            return None

        return "0x%02x" % addr


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        int_str = jdict.get('int')
        ext_str = jdict.get('ext')

        int_addr = None if int_str is None else int(int_str, 0)
        ext_addr = None if ext_str is None else int(ext_str, 0)

        return cls(int_addr, ext_addr)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, int_addr, ext_addr):
        """
        Constructor
        """
        super().__init__()

        self.__int_addr = int_addr          # int       I2C address of SHT in A4 package
        self.__ext_addr = ext_addr          # int       I2C address of SHT exposed to air


    def __eq__(self, other):
        try:
            return self.int_addr == other.int_addr and self.ext_addr == other.ext_addr

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic
    def int_sht(self):
        return None


    # noinspection PyMethodMayBeStatic
    def ext_sht(self):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def int_addr(self):
        return self.__int_addr


    @property
    def ext_addr(self):
        return self.__ext_addr


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['int'] = SHTConf.__addr_str(self.__int_addr)
        jdict['ext'] = SHTConf.__addr_str(self.__ext_addr)

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SHTConf(core):{int_addr:%s, ext_addr:%s}" %  \
               (SHTConf.__addr_str(self.int_addr), SHTConf.__addr_str(self.ext_addr))



"""
Created on 9 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The A4CalibratedDatum is a normalised electrochem output voltage. Its vCal field represents the calibrated weV
minus the calibrated aeV.

The vXCal field should only be computed when the NO2 concentration is known precisely, so this should be done
in the preprocessing phase for the O3 model.

example document:
{"weV": 0.41138, "aeV": 0.40257, "weC": 0.02671, "cnc": 116.5, "vCal": 17.292, "vXCal": 0.00212}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum

from scs_core.gas.a4.a4_calib import A4Calib
from scs_core.gas.a4.a4_datum import A4Datum


# --------------------------------------------------------------------------------------------------------------------

class A4Calibrator(object):
    """
    classdocs
    """

    __CALIBRATED_GASES = ('CO', 'H2S', 'NO', 'NO2', 'Ox', 'SO2', 'VOCe')

    @classmethod
    def calibrated_gases(cls):
        return cls.__CALIBRATED_GASES


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calib: A4Calib):
        """
        Constructor
        """
        self.__we_elc_v = round(calib.we_elc_mv / 1000.0, 3)                        # we_electronic_zero_mv
        self.__ae_elc_v = round(calib.ae_elc_mv / 1000.0, 3)                        # ae_electronic_zero_mv

        self.__we_sens_v = round(calib.we_sens_mv / 1000.0, 6)                      # we_sensitivity_mv_ppb

        if calib.we_no2_x_sens_mv is None:
            self.__we_no2_x_sens_v = None
        else:
            self.__we_no2_x_sens_v = round(calib.we_no2_x_sens_mv / 1000.0, 6)      # we_cross_sensitivity_no2_mv_ppb


    # ----------------------------------------------------------------------------------------------------------------

    def calibrate(self, datum):
        # zero offset...
        we_v_zero_cal = datum.we_v - self.we_elc_v
        ae_v_zero_cal = datum.ae_v - self.ae_elc_v

        # remove noise...
        v_zero_cal = we_v_zero_cal - ae_v_zero_cal

        # gain...
        v_cal = v_zero_cal / self.we_sens_v

        return A4CalibratedDatum(datum.we_v, datum.ae_v, datum.we_c, datum.cnc, v_cal)


    def set_we_v_zero_x_cal(self, datum, no2_cnc):
        we_v_x = no2_cnc * self.we_no2_x_sens_v
        datum.we_v_zero_x_cal = we_v_x - self.we_elc_v


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_elc_v(self):
        return self.__we_elc_v


    @property
    def ae_elc_v(self):
        return self.__ae_elc_v


    @property
    def we_sens_v(self):
        return self.__we_sens_v


    @property
    def we_no2_x_sens_v(self):
        return self.__we_no2_x_sens_v


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Calibrator:{we_elc_v:%s, ae_elc_v:%s, we_sens_v:%s, we_no2_x_sens_v:%s}" % \
               (self.we_elc_v, self.ae_elc_v, self.we_sens_v, self.we_no2_x_sens_v)


# --------------------------------------------------------------------------------------------------------------------

class A4CalibratedDatum(A4Datum):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        we_v = jdict.get('weV')
        ae_v = jdict.get('aeV')

        we_c = jdict.get('weC')
        cnc = jdict.get('cnc')

        v_cal = jdict.get('vCal')
        we_v_zero_x_cal = jdict.get('xCal')

        return cls(we_v, ae_v, we_c, cnc, v_cal, we_v_zero_x_cal)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, ae_v, we_c, cnc, v_cal, we_v_zero_x_cal=None):
        """
        Constructor
        """
        super().__init__(we_v, ae_v, we_c, cnc)

        self.__v_cal = Datum.float(v_cal, 3)                            # calibrated voltage (ppb)
        self.__we_v_zero_x_cal = Datum.float(we_v_zero_x_cal, 6)        # response to NO2 cross-sensitivity (Volts)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v
        jdict['aeV'] = self.ae_v

        jdict['weC'] = self.we_c                                # may be None
        jdict['cnc'] = self.cnc                                 # may be None

        if self.v_cal is not None:
            jdict['vCal'] = self.v_cal

        if self.we_v_zero_x_cal is not None:
            jdict['xCal'] = self.we_v_zero_x_cal

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def v_cal(self):
        return self.__v_cal


    @property
    def we_v_zero_x_cal(self):
        return self.__we_v_zero_x_cal


    @we_v_zero_x_cal.setter
    def we_v_zero_x_cal(self, we_v_zero_x_cal):
        self.__we_v_zero_x_cal = round(we_v_zero_x_cal, 9)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4CalibratedDatum:{we_v:%s, ae_v:%s, we_c:%s, cnc:%s, v_cal:%s, we_v_zero_x_cal:%s}" % \
               (self.we_v, self.ae_v, self.we_c, self.cnc, self.v_cal, self.we_v_zero_x_cal)



"""
Created on 17 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alert example:
{"id": 77, "topic": "south-coast-science-dev/development/loc/1/gases", "field": "val.CO.cnc", "lower-threshold": null,
"upper-threshold": 1000.0, "alert-on-none": true, "aggregation-period": {"interval": 1, "units": "M"},
"test-interval": null, "creator-email-address": "authorization@southcoastscience.com",
"to": "someone@me.com", "cc-list": [], "suspended": false}

AlertStatus example:
{"id": 77, "rec": "2021-09-07T11:40:00Z", "cause": null, "val": 589.6}

https://martinstapel.com/how-to-autoincrement-in-dynamo-db-if-you-really-need-to/
https://stackoverflow.com/questions/37072341/how-to-use-auto-increment-for-primary-key-id-in-dynamodb
"""

from collections import OrderedDict

from scs_core.data.recurring_period import RecurringPeriod
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class AlertStatus(JSONable):
    """
    classdocs
    """

    BELOW_LOWER_THRESHOLD =     '<L'
    ABOVE_UPPER_THRESHOLD =     '>U'
    NULL_VALUE =                'NV'

    @classmethod
    def causes(cls):
        return [cls.BELOW_LOWER_THRESHOLD, cls.ABOVE_UPPER_THRESHOLD, cls.NULL_VALUE]


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')

        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        cause = jdict.get('cause')
        value = jdict.get('val')

        return cls(id, rec, cause, value)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, rec, cause, value):
        """
        Constructor
        """
        self.__id = Datum.int(id)                                   # int

        self.__rec = rec                                            # LocalizedDatetime
        self.__cause = cause                                        # string
        self.__value = Datum.float(value)                           # float


    def __lt__(self, other):
        if self.id < other.id:
            return True

        if self.id > other.id:
            return False

        if self.rec < other.rec:
            return True

        if self.rec > other.rec:
            return False

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def is_excursion(self):
        return self.cause is not None


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id

        jdict['rec'] = self.rec
        jdict['cause'] = self.cause
        jdict['val'] = self.value

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def rec(self):
        return self.__rec


    @property
    def cause(self):
        return self.__cause


    @property
    def value(self):
        return self.__value


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertStatus:{id:%s, rec:%s, cause:%s, value:%s}" %  \
               (self.id, self.rec, self.cause, self.value)


# --------------------------------------------------------------------------------------------------------------------

class AlertSpecification(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')

        topic = jdict.get('topic')
        field = jdict.get('field')

        lower_threshold = jdict.get('lower-threshold')
        upper_threshold = jdict.get('upper-threshold')
        alert_on_none = jdict.get('alert-on-none')

        aggregation_period = RecurringPeriod.construct_from_jdict(jdict.get('aggregation-period'))
        test_interval = RecurringPeriod.construct_from_jdict(jdict.get('test-interval'))

        creator_email_address = jdict.get('creator-email-address')

        to = jdict.get('to')
        cc_list = jdict.get('cc-list')
        suspended = jdict.get('suspended')

        return cls(id, topic, field, lower_threshold, upper_threshold, alert_on_none,
                   aggregation_period, test_interval, creator_email_address, to, cc_list, suspended)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, topic, field, lower_threshold, upper_threshold, alert_on_none,
                 aggregation_period, test_interval, creator_email_address, to, cc_list, suspended):
        """
        Constructor
        """
        self.__id = Datum.int(id)                                   # int

        self.__topic = topic                                        # string topic
        self.__field = field                                        # string path

        self.__lower_threshold = Datum.float(lower_threshold)       # float                 updatable
        self.__upper_threshold = Datum.float(upper_threshold)       # float                 updatable
        self.__alert_on_none = bool(alert_on_none)                  # bool                  updatable

        self.__aggregation_period = aggregation_period              # RecurringPeriod       updatable
        self.__test_interval = test_interval                        # RecurringPeriod       updatable

        self.__creator_email_address = creator_email_address        # string

        self.__to = to                                              # string                updatable
        self.__cc_list = cc_list                                    # array of string       updatable
        self.__suspended = bool(suspended)                          # bool                  updatable


    def __lt__(self, other):
        if self.topic < other.topic:
            return True

        if self.topic > other.topic:
            return False

        if self.field < other.field:
            return True

        if self.field > other.field:
            return False

        if self.creator_email_address < other.creator_email_address:
            return True

        if self.creator_email_address > other.creator_email_address:
            return False

        if self.to < other.to:
            return True

        if self.to > other.to:
            return False

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.topic is None or self.field is None or self.aggregation_period is None or \
                self.creator_email_address is None or self.to is None:
            return False

        if not self.has_trigger():
            return False

        if not self.has_valid_thresholds():
            return False

        # if self.test_interval is not None and not self.test_interval < self.aggregation_period:
        #     return False

        return True


    def has_trigger(self):
        return self.alert_on_none or self.lower_threshold is not None or self.upper_threshold is not None


    def has_valid_thresholds(self):
        if self.lower_threshold is None or self.upper_threshold is None:
            return True

        return self.lower_threshold < self.upper_threshold


    def has_valid_aggregation_period(self):
        return self.aggregation_period is not None and self.aggregation_period.is_valid()


    def may_update(self, other):
        if self.id != other.id or self.topic != other.topic or self.field != other.field or \
                self.creator_email_address != other.creator_email_address:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def status(self, value):
        if self.suspended:
            return None

        if value is None and self.alert_on_none:
            cause = AlertStatus.NULL_VALUE

        elif self.lower_threshold is not None and value is not None and value < self.lower_threshold:
            cause = AlertStatus.BELOW_LOWER_THRESHOLD

        elif self.upper_threshold is not None and value is not None and value > self.upper_threshold:
            cause = AlertStatus.ABOVE_UPPER_THRESHOLD

        else:
            cause = None

        return AlertStatus(self.id, LocalizedDatetime.now().utc(), cause, value)


    # ----------------------------------------------------------------------------------------------------------------

    def checkpoint(self):     # TODO: needs an offset if test_interval is used
        return self.aggregation_period.checkpoint()


    def cron(self, minutes_offset):
        return self.aggregation_period.cron(minutes_offset)


    def aws_cron(self, minutes_offset):
        return self.aggregation_period.aws_cron(minutes_offset)
        # return self.test_interval.cron(minutes_offset) if self.test_interval else \
        #     self.aggregation_period.cron(minutes_offset)


    def timedelta(self):
        return self.aggregation_period.timedelta()


    def end_datetime(self, origin: LocalizedDatetime):
        return self.aggregation_period.end_datetime(origin)
        # return self.test_interval.end_datetime(origin) if self.test_interval else \
        #     self.aggregation_period.end_datetime(origin)

    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id

        jdict['topic'] = self.topic
        jdict['field'] = self.field

        jdict['lower-threshold'] = self.lower_threshold
        jdict['upper-threshold'] = self.upper_threshold
        jdict['alert-on-none'] = self.alert_on_none

        jdict['aggregation-period'] = self.aggregation_period.as_json()
        jdict['test-interval'] = None if self.test_interval is None else self.test_interval.as_json()

        jdict['creator-email-address'] = self.creator_email_address

        jdict['to'] = self.to
        jdict['cc-list'] = self.cc_list
        jdict['suspended'] = self.suspended

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def topic(self):
        return self.__topic


    @property
    def field(self):
        return self.__field


    @property
    def lower_threshold(self):
        return self.__lower_threshold


    @property
    def upper_threshold(self):
        return self.__upper_threshold


    @property
    def alert_on_none(self):
        return self.__alert_on_none


    @property
    def aggregation_period(self):
        return self.__aggregation_period


    @property
    def test_interval(self):
        return self.__test_interval


    @property
    def creator_email_address(self):
        return self.__creator_email_address


    @property
    def to(self):
        return self.__to


    @property
    def cc_list(self):
        return self.__cc_list


    @property
    def suspended(self):
        return self.__suspended


    # ----------------------------------------------------------------------------------------------------------------

    @id.setter
    def id(self, id):
        self.__id = id


    def append_to_cc_list(self, cc):
        if cc in self.__cc_list:
            return

        self.__cc_list.append(cc)


    def remove_from_cc_list(self, cc):
        if cc not in self.__cc_list:
            return

        self.__cc_list.remove(cc)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertSpecification:{id:%s, topic:%s, field:%s, lower_threshold:%s, upper_threshold:%s, " \
               "alert_on_none:%s, aggregation_period:%s, test_interval:%s, creator_email_address:%s, " \
               "to:%s, cc_list:%s, suspended:%s}" %  \
               (self.id, self.topic, self.field, self.lower_threshold, self.upper_threshold,
                self.alert_on_none, self.aggregation_period, self.test_interval, self.creator_email_address,
                self.to, self.cc_list, self.suspended)



"""
Created on 30 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

GNSS DOP and Active Satellites
$xxGSA,opMode,navMode{,sv},PDOP,HDOP,VDOP*cs

example sentence:
$GPGSA,A,3,23,29,07,08,09,18,26,28,,,,,1.94,1.18,1.54*0D

example values:
GPGSA:{op_mode:A, nav_mode:3, sv:[21, 02, 28, 13, 30, 05, None, None, None, None, None, None], pdop:4.61,
hdop:3.10, vdop:3.41}

GPGSA:{op_mode:A, nav_mode:1, sv:[None, None, None, None, None, None, None, None, None, None, None, None], pdop:99.99,
hdop:99.99, vdop:99.99}

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183
"""

from scs_core.data.str import Str
from scs_core.position.nmea.nmea_sentence import NMEASentence


# --------------------------------------------------------------------------------------------------------------------

class GPGSA(NMEASentence):
    """
    classdocs
    """

    MESSAGE_IDS = ("$GNGSA", "$GPGSA")

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, r):
        if r.message_id not in cls.MESSAGE_IDS:
            raise TypeError("invalid sentence:%s" % r)

        op_mode = r.str(1)
        nav_mode = r.int(2)

        sv = [r.int(i + 3) for i in range(12)]

        pdop = r.float(15, 2)
        hdop = r.float(16, 2)
        vdop = r.float(17, 2)

        return GPGSA(r.message_id, op_mode, nav_mode, sv, pdop, hdop, vdop)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, message_id, op_mode, nav_mode, sv, pdop, hdop, vdop):
        """
        Constructor
        """
        super().__init__(message_id)

        self.__op_mode = op_mode            # string
        self.__nav_mode = nav_mode          # int

        self.__sv = sv                      # list of int

        self.__pdop = pdop                  # float(2)
        self.__hdop = hdop                  # float(2)
        self.__vdop = vdop                  # float(2)


    def __eq__(self, other):
        try:
            return self.op_mode == other.op_mode and self.nav_mode == other.nav_mode and self.sv == other.sv and \
                   self.pdop == other.pdop and self.hdop == other.hdop and self.vdop == other.vdop

        except AttributeError:
            return False


    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return not self == other


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def op_mode(self):
        return self.__op_mode


    @property
    def nav_mode(self):
        return self.__nav_mode


    @property
    def sv(self):
        return self.__sv


    @property
    def pdop(self):
        return self.__pdop


    @property
    def hdop(self):
        return self.__hdop


    @property
    def vdop(self):
        return self.__vdop


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPGSA:{source:%s, op_mode:%s, nav_mode:%s, sv:%s, pdop:%s, hdop:%s, vdop:%s}" % \
               (self.source, self.op_mode, self.nav_mode, Str.collection(self.__sv), self.pdop, self.hdop, self.vdop)



"""
Created on 15 Oct 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A catalogue entry for a group of machine learning model compendia, one for each gas

By convention, the name of the group should be the same as the name of the ML model template.

document example:
{"name": "oE.1", "compendia": {"NO2": "NO2.vE.OPCube.21HA"}}
"""

import os

from collections import OrderedDict

from scs_core.data.json import JSONCatalogueEntry
from scs_core.data.path_dict import PathDict
from scs_core.data.str import Str

from scs_core.model.gas.baseline import Baseline
from scs_core.model.catalogue.model_compendium import ModelCompendium

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModelCompendiumGroup(JSONCatalogueEntry):
    """
    classdocs
    """

    __CATALOGUE_NAME = 'groups'

    @classmethod
    def catalogue_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), cls.__CATALOGUE_NAME)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, {}) if skeleton else None

        name = jdict.get('name')
        compendia_jdict = jdict.get('compendia')

        compendia = {gas: ModelCompendium.retrieve(compendium_name) for gas, compendium_name in compendia_jdict.items()}

        return cls(name, compendia)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, compendia):
        """
        Constructor
        """
        self.__name = name
        self.__compendia = compendia                    # dict of gas: ModelCompendium

        self.__logger = Logging.getLogger()


    def __len__(self):
        return len(self.compendia)


    # ----------------------------------------------------------------------------------------------------------------

    def preprocess(self, datum: PathDict, offsets=None):
        if offsets is None:
            offsets = Baseline({})

        for gas, compendium in self.__compendia.items():
            datum = compendium.preprocess(datum, offset=offsets.sensor_offset(gas))

        return datum


    def postprocess(self, preprocessed: PathDict, response: PathDict):
        if not response.has_sub_path(sub_path='exg'):
            return None

        for gas, compendium in self.__compendia.items():
            try:
                vcal_excess_path = '.'.join(('sample', 'val', gas, 'vCalExtr'))
                vcal_excess = float(preprocessed.node(vcal_excess_path))

                model_output_path = '.'.join(('exg', 'val', gas, 'cnc'))
                model_output = float(response.node(model_output_path))

            except KeyError:
                continue

            for primary in compendium.primaries:
                corrected_exg = round(compendium.postprocess(primary, vcal_excess, model_output), 1)

                self.__logger.debug("postprocess - %s model_output: %s vcal_excess: %s corrected_exg: %s" %
                                    (gas, model_output, vcal_excess, corrected_exg))

                response.append(model_output_path, corrected_exg)

        return response.node('exg')


    # ----------------------------------------------------------------------------------------------------------------

    def add(self, gas, compendium):
        self.__compendia[gas] = compendium


    def exclude(self, gas):
        if gas in self.compendia:
            del self.__compendia[gas]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    @property
    def gases(self):
        return sorted(self.compendia.keys())


    def compendium(self, gas):
        return self.__compendia[gas]


    @property
    def compendia(self):
        return self.__compendia


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['name'] = self.name
        jdict['compendia'] = {gas: self.compendia[gas].name for gas in self.gases}

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        compendia = Str.collection(self.compendia)

        return "ModelCompendiumGroup:{name:%s, compendia:%s}" %  (self.name, compendia)



"""
Created on 21 Jun 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

specifies which PSU is present, if any

example JSON:
{"model": "MobileV2", "batt-model": "PackV1", "ignore-threshold": true, "reporting-interval": 10,
"report-file": "/tmp/southcoastscience/psu_status_report.json"}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class PSUConf(PersistentJSONable):
    """
    classdocs
    """

    __DEFAULT_REPORTING_INTERVAL = 10               # seconds

    __FILENAME = "psu_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        psu_model = jdict.get('model')
        batt_model = jdict.get('batt-model')
        ignore_threshold = jdict.get('ignore-threshold', False)

        reporting_interval = jdict.get('reporting-interval', cls.__DEFAULT_REPORTING_INTERVAL)
        report_file = jdict.get('report-file')

        return cls(psu_model, batt_model, ignore_threshold, reporting_interval, report_file)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, psu_model, batt_model, ignore_threshold, reporting_interval, report_file):
        """
        Constructor
        """
        super().__init__()

        self.__psu_model = psu_model                                        # string
        self.__batt_model = batt_model                                      # string
        self.__ignore_threshold = ignore_threshold                          # bool

        self.__reporting_interval = int(reporting_interval)                 # int
        self.__report_file = report_file                                    # string


    def __eq__(self, other):
        try:
            return self.psu_model == other.psu_model and \
                   self.batt_model == other.batt_model and \
                   self.ignore_threshold == other.ignore_threshold and \
                   self.reporting_interval == other.reporting_interval and \
                   self.report_file == other.report_file

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def psu(self, host, interface_model):
        return None


    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def psu_monitor(self, host, interface_model, ignore_standby):
        return None


    # noinspection PyMethodMayBeStatic
    def psu_class(self):
        return None


    # noinspection PyMethodMayBeStatic
    def psu_report_class(self):
        return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def psu_model(self):
        return self.__psu_model


    @property
    def batt_model(self):
        return self.__batt_model


    @property
    def ignore_threshold(self):
        return self.__ignore_threshold


    @property
    def report_file(self):
        return self.__report_file


    @property
    def reporting_interval(self):
        return self.__reporting_interval


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['model'] = self.__psu_model
        jdict['batt-model'] = self.__batt_model
        jdict['ignore-threshold'] = self.__ignore_threshold

        jdict['reporting-interval'] = self.__reporting_interval
        jdict['report-file'] = self.__report_file

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PSUConf(core):{psu_model:%s, batt_model:%s, ignore_threshold:%s, reporting_interval:%s, " \
               "report_file:%s}" % \
               (self.psu_model, self.batt_model, self.ignore_threshold, self.reporting_interval,
                self.report_file)



"""
Created on 21 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from scs_core.osio.client.rest_client import RESTClient
from scs_core.osio.data.user import User
from scs_core.osio.data.user_metadata import UserMetadata


# --------------------------------------------------------------------------------------------------------------------

class UserManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, user_id):
        request_path = '/v1/users/' + user_id

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.get(request_path)
        except RuntimeError:
            response_jdict = None

        self.__rest_client.close()

        user = User.construct_from_jdict(response_jdict)

        return user


    def find_public(self, user_id):
        request_path = '/v1/public/users/' + user_id

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.get(request_path)
        except RuntimeError:
            response_jdict = None

        self.__rest_client.close()

        user = UserMetadata.construct_from_jdict(response_jdict)

        return user


    def find_members_of_org(self, org_id):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def update(self, user_id, user):
        request_path = '/v1/users/' + user_id

        # request...
        self.__rest_client.connect()

        try:
            self.__rest_client.put(request_path, user.as_json())
        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "UserManager:{rest_client:%s}" % self.__rest_client



"""
Created on 15 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import boto3

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.config.aws import AWS


# --------------------------------------------------------------------------------------------------------------------

class Client(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, category, key: AccessKey):
        if key.ok():
            return boto3.client(
                category,
                aws_access_key_id=key.id,
                aws_secret_access_key=key.secret,
                region_name=AWS.region()
            )

        return boto3.client(category, region_name=AWS.region())


    @classmethod
    def resource(cls, category, key: AccessKey):
        if key.ok():
            return boto3.resource(
                category,
                aws_access_key_id=key.id,
                aws_secret_access_key=key.secret,
                region_name=AWS.region()
            )

        return boto3.resource(category, region_name=AWS.region())



"""
Created on 1 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"sn1": {"calibrated-on": "2019-02-02T11:34:16Z", "offset": 50, "env": {"hmd": 66.0, "tmp": 11.0, "pA": 99.0}},
"sn2": {"calibrated-on": "2019-02-02T11:30:17Z", "offset": 0, "env": null},
"sn3": {"calibrated-on": "2019-02-02T11:30:17Z", "offset": 0, "env": null},
"sn4": {"calibrated-on": "2019-02-02T11:30:17Z", "offset": 0, "env": null}}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable
from scs_core.data.str import Str

from scs_core.gas.sensor_baseline import SensorBaseline


# --------------------------------------------------------------------------------------------------------------------

class AFEBaseline(PersistentJSONable):
    """
    classdocs
    """

    __SENSORS = 4       # TODO: better to find out how long the AFECalib is than to use a constant

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME =    "afe_baseline.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls.null_datum() if skeleton else None

        sensor_baselines = []

        for i in range(len(jdict)):
            key = 'sn' + str(i + 1)

            base = SensorBaseline.construct_from_jdict(jdict[key]) if key in jdict else SensorBaseline(None, 0)
            sensor_baselines.append(base)

        return cls(sensor_baselines)


    @classmethod
    def null_datum(cls):
        return cls([SensorBaseline(None, 0)] * cls.__SENSORS)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_baselines):
        """
        Constructor
        """
        super().__init__()

        self.__sensor_baselines = sensor_baselines        # array of SensorBaseline


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for i in range(len(self)):
                if self.sensor_baseline(i) != other.sensor_baseline(i):
                    return False

                return True

        except (TypeError, AttributeError):
            return False


    def __len__(self):
        return len(self.__sensor_baselines)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        for i in range(len(self.__sensor_baselines)):
            jdict['sn' + str(i + 1)] = self.__sensor_baselines[i]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def sensor_baseline(self, i):
        return self.__sensor_baselines[i]


    def set_sensor_baseline(self, i, sensor_baseline):
        self.__sensor_baselines[i] = sensor_baseline


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AFEBaseline:{sensor_baselines:%s}" % Str.collection(self.__sensor_baselines)



"""
Created on 6 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

AQCSV: Methods

NB: initialisation is performed at the foot of this class

example:
{"Parameter": "Oxides of nitrogen (NOx)", "Parameter Code": 42603, "Method Code": 31, "Recording Mode": "Continuous",
"Collection Description": "INSTRUMENTAL", "Analysis Description": "CHEMILUMINESCENCE", "Method Type": "FRM",
"Reference Method ID": "RFNA-1078-031", "Equivalent Method": "MELOY NA530R", "Federal MDL": "10",
"Min Value": "-5", "Max Value": "1200", "Digits": "1", "Round Truncate Indicator": "T",
"Units": "Parts per billion"}

suspect rows:
Outdoor Temperature,62101,59,Continuous,Instrumental,Vaisala HMP 155,,,,-60,,,1,R,Degrees Fahrenheit
Outdoor Temperature,62101,59,Continuous,Instrumental,Vaisala HMP 155,,,,-60,,,1,R,Degrees Fahrenheit

https://www.airnow.gov/
"""

import os

from collections import OrderedDict

from scs_core.csv.csv_archive import CSVArchive

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Method(CSVArchive, JSONable):
    """
    classdocs
    """

    _retrieved = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def archive_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'archive', 'methods.csv')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def find_by_parameter_code(cls, parameter_code):
        for method in cls._retrieved.values():
            if method.parameter_code == parameter_code:
                yield method


    @classmethod
    def find_by_method_code(cls, method_code):
        for method in cls._retrieved.values():
            if method.method_code == method_code:
                yield method


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        parameter = jdict.get('Parameter')
        parameter_code = jdict.get('Parameter Code')
        method_code = jdict.get('Method Code')
        recording_mode = jdict.get('Recording Mode')

        collection_description = jdict.get('Collection Description')
        analysis_description = jdict.get('Analysis Description')

        method_type = jdict.get('Method Type')
        reference_method_id = jdict.get('Reference Method ID')
        equivalent_method = jdict.get('Equivalent Method')

        federal_mdl = jdict.get('Federal MDL')

        min_value = jdict.get('Min Value')
        max_value = jdict.get('Max Value')
        digits = jdict.get('Digits')

        round_truncate_indicator = jdict.get('Round Truncate Indicator')
        units = jdict.get('Units')

        return Method(parameter, parameter_code, method_code, recording_mode,
                      collection_description, analysis_description, method_type, reference_method_id, equivalent_method,
                      federal_mdl, min_value, max_value, digits, round_truncate_indicator, units)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, parameter, parameter_code, method_code, recording_mode,
                 collection_description, analysis_description, method_type, reference_method_id, equivalent_method,
                 federal_mdl, min_value, max_value, digits, round_truncate_indicator, units):
        """
        Constructor
        """
        self.__parameter = parameter                                        # string
        self.__parameter_code = int(parameter_code)                         # int(5)
        self.__method_code = int(method_code)                               # int(3)
        self.__recording_mode = recording_mode                              # string

        self.__collection_description = collection_description              # string
        self.__analysis_description = analysis_description                  # string
        self.__method_type = method_type                                    # string
        self.__reference_method_id = reference_method_id                    # string
        self.__equivalent_method = equivalent_method                        # string

        self.__federal_mdl = Datum.float(federal_mdl)                       # float
        self.__min_value = Datum.float(min_value)                           # float
        self.__max_value = Datum.float(max_value)                           # float
        self.__digits = Datum.int(digits)                                   # int

        self.__round_truncate_indicator = round_truncate_indicator          # string
        self.__units = units                                                # string


    def __eq__(self, other):
        try:
            return self.parameter == other.parameter and \
                   self.parameter_code == other.parameter_code and \
                   self.method_code == other.method_code and \
                   self.recording_mode == other.recording_mode and \
                   self.collection_description == other.collection_description and \
                   self.analysis_description == other.analysis_description and \
                   self.method_type == other.method_type and \
                   self.reference_method_id == other.reference_method_id and \
                   self.equivalent_method == other.equivalent_method and \
                   self.federal_mdl == other.federal_mdl and \
                   self.min_value == other.min_value and \
                   self.max_value == other.max_value and \
                   self.digits == other.digits and \
                   self.round_truncate_indicator == other.round_truncate_indicator and \
                   self.units == other.units

        except AttributeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['Parameter'] = self.parameter
        jdict['Parameter Code'] = self.parameter_code
        jdict['Method Code'] = self.method_code
        jdict['Recording Mode'] = self.recording_mode

        jdict['Collection Description'] = self.collection_description
        jdict['Analysis Description'] = self.analysis_description
        jdict['Method Type'] = self.method_type
        jdict['Reference Method ID'] = self.reference_method_id
        jdict['Equivalent Method'] = self.equivalent_method

        jdict['Federal MDL'] = self.federal_mdl
        jdict['Min Value'] = self.min_value
        jdict['Max Value'] = self.max_value
        jdict['Digits'] = self.digits

        jdict['Round Truncate Indicator'] = self.round_truncate_indicator
        jdict['Units'] = self.units

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.parameter_code, self.method_code


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def parameter(self):
        return self.__parameter


    @property
    def parameter_code(self):
        return self.__parameter_code


    @property
    def method_code(self):
        return self.__method_code


    @property
    def recording_mode(self):
        return self.__recording_mode


    @property
    def collection_description(self):
        return self.__collection_description


    @property
    def analysis_description(self):
        return self.__analysis_description


    @property
    def method_type(self):
        return self.__method_type


    @property
    def reference_method_id(self):
        return self.__reference_method_id


    @property
    def equivalent_method(self):
        return self.__equivalent_method


    @property
    def federal_mdl(self):
        return self.__federal_mdl


    @property
    def min_value(self):
        return self.__min_value


    @property
    def max_value(self):
        return self.__max_value


    @property
    def digits(self):
        return self.__digits


    @property
    def round_truncate_indicator(self):
        return self.__round_truncate_indicator


    @property
    def units(self):
        return self.__units


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Method:{parameter:%s, parameter_code:%05d, method_code:%03d, recording_mode:%s, " \
               "collection_description:%s, analysis_description:%s, method_type:%s, reference_method_id:%s, " \
               "equivalent_method:%s, federal_mdl:%s, min_value:%s, max_value:%s, digits:%s, " \
               "round_truncate_indicator:%s, units:%s}" % \
               (self.parameter, self.parameter_code, self.method_code, self.recording_mode,
                self.collection_description, self.analysis_description, self.method_type, self.reference_method_id,
                self.equivalent_method, self.federal_mdl, self.min_value, self.max_value, self.digits,
                self.round_truncate_indicator, self.units)


# --------------------------------------------------------------------------------------------------------------------
# initialisation...

Method.retrieve()



"""
Created on 09 Oct 2020

@author: Jade Page (jade.page@southcoastscience.com)
"""

import json
import os
import shutil

from collections import OrderedDict
from urllib.request import urlopen

from scs_core.aws.config.aws import AWS
from scs_core.data.json import PersistentJSONable
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class AWSIdentity(PersistentJSONable):
    """
    classdocs
    """

    __KEEP_ALIVE = 60               # seconds

    __CERTS_PATH = "/greengrass/certs/"
    __ATS_ROOT_CA_RSA_2048_REMOTE_LOCATION = "https://www.amazontrust.com/repository/AmazonRootCA1.pem"

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "greengrass_identity.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if jdict is None:
            return None

        core_name = jdict.get('core-name')
        group_name = jdict.get('group-name')
        iot_client = None
        gg_client = None

        return AWSIdentity(iot_client, gg_client, core_name, group_name)


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def return_default_policy():
        # create iot policy

        core_policy_doc = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    # iot data plane
                    "Action": ["iot:Publish", "iot:Subscribe", "iot:Connect", "iot:Receive", "iot:GetThingShadow",
                               "iot:DeleteThingShadow", "iot:UpdateThingShadow"],
                    "Resource": ["arn:aws:iot:us-west-2:*:*"]
                },
                {
                    "Effect": "Allow",
                    # Greengrass data plane
                    "Action": ["greengrass:AssumeRoleForGroup", "greengrass:CreateCertificate",
                               "greengrass:GetConnectivityInfo", "greengrass:GetDeployment",
                               "greengrass:GetDeploymentArtifacts", "greengrass:UpdateConnectivityInfo",
                               "greengrass:UpdateCoreDeploymentStatus"],
                    "Resource": ["*"]
                }
            ]
        }
        return core_policy_doc


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, iot_client, gg_client, core_name, group_name):
        """
        Constructor
        """
        super().__init__()

        self.__logger = Logging.getLogger()

        self.__iot_client = iot_client
        self.__gg_client = gg_client
        self.__core_name = core_name
        self.__group_name = group_name
        self.__thing_arn = None
        self.__thing_id = None
        self.__certificate = None
        self.__latest_core_version_arn = None
        self.__latest_logger_version_arn = None
        self.__latest_group_version_arn = None
        self.__hash = None


    def __eq__(self, other):
        try:
            return self.core_name == other.core_name and self.group_name == other.group_name

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def setup_device(self):
        self.create_logger()
        self.create_thing()
        self.create_core()
        self.create_group()
        self.persist_certs()
        self.update_config_file()


    # ----------------------------------------------------------------------------------------------------------------

    def create_thing(self):
        # Create core
        res = self.__iot_client.create_thing(thingName=self.__core_name)
        self.__thing_arn = res["thingArn"]
        self.__thing_id = res["thingId"]

        self.__logger.info("Core created")

        # Create certs
        self.__certificate = self.__iot_client.create_keys_and_certificate(setAsActive=True)

        self.__logger.info("Cert created")

        # Attach cert to core
        cert_arn = self.__certificate["certificateArn"]
        self.__iot_client.attach_thing_principal(
            thingName=self.__core_name,
            principal=cert_arn,
        )

        self.__logger.info("Cert attached to core")

        # Create policy
        policy_doc = self.return_default_policy()
        policy_name = "{}_basic_policy".format(self.__core_name)

        self.__iot_client.create_policy(
            policyName=policy_name,
            policyDocument=json.dumps(policy_doc)
        )

        self.__logger.info("Policy created")
        self.__logger.info(json.dumps(policy_doc))

        # Attach policy
        # associate iot policy with core cert
        self.__iot_client.attach_policy(
            policyName=policy_name,
            target=cert_arn,
        )

        self.__logger.info("Policy attached to cert")


    def create_core(self):
        initial_core_definition_version = {
            'Cores': [
                {
                    'Id': self.__core_name,
                    'CertificateArn': self.__certificate["certificateArn"],
                    'SyncShadow': False,
                    'ThingArn': self.__thing_arn,
                }
            ]
        }
        res = self.__gg_client.create_core_definition(
            Name="{}_def".format(self.__core_name),
            InitialVersion=initial_core_definition_version,
        )
        self.__latest_core_version_arn = res['LatestVersionArn']

        self.__logger.info("Core created")
        self.__logger.info(initial_core_definition_version)


    def create_group(self):
        initial_group_definition_version = {
            'CoreDefinitionVersionArn': self.__latest_core_version_arn,
            'LoggerDefinitionVersionArn': self.__latest_logger_version_arn,
        }

        res = self.__gg_client.create_group(
            InitialVersion=initial_group_definition_version,
            Name=self.__group_name,
        )
        self.__latest_group_version_arn = res['LatestVersionArn']

        self.__logger.info("Group definition created")
        self.__logger.info(initial_group_definition_version)


    def create_logger(self):
        res = self.__gg_client.create_logger_definition(
            InitialVersion={
                'Loggers': [
                    {
                        'Component': 'GreengrassSystem',
                        'Id': 'Logger_definition_to_greengrass_system_' + self.__group_name,
                        'Level': 'INFO',
                        'Space': 1280,
                        'Type': 'FileSystem',
                    },
                    {
                        'Component': 'Lambda',
                        'Id': 'Logger_definition_to_lambda_' + self.__group_name,
                        'Level': 'INFO',
                        'Space': 1280,
                        'Type': 'FileSystem',
                    },
                ]
            },
            Name='Logger_definition_' + self.__group_name,
        )
        self.__latest_logger_version_arn = res['LatestVersionArn']

        self.__logger.info("Logger created")


    def persist_certs(self):
        # Delete existing keys from dir
        # Add new keys
        keys = self.__certificate["keyPair"]
        self.__hash = self.__certificate["certificateId"]
        self.__hash = self.__hash[0: 10]
        certificate_path = self.__hash + ".cert.pem"
        private_path = self.__hash + ".private.key"
        public_path = self.__hash + ".public.key"

        certs_dir = self.__CERTS_PATH
        if os.path.isdir(certs_dir):
            shutil.rmtree(certs_dir)  # remove dir and all contains
            self.__logger.info("Removed old keys")

        os.mkdir(certs_dir)

        data = urlopen(url=self.__ATS_ROOT_CA_RSA_2048_REMOTE_LOCATION)
        with open(certs_dir + "root.ca.pem", "wb") as local_file:
            local_file.write(data.read())

        f = open(certs_dir + "/" + certificate_path, "w")
        f.write(self.__certificate["certificatePem"])
        f.close()

        self.__logger.info("Cert saved")

        f = open(certs_dir + "/" + private_path, "w")
        f.write(keys["PrivateKey"])
        f.close()

        self.__logger.info("Private key saved")

        f = open(certs_dir + "/" + public_path, "w")
        f.write(keys["PublicKey"])
        f.close()

        self.__logger.info("Public key saved")


    def update_config_file(self):
        res = self.__iot_client.describe_endpoint(
            endpointType='iot:Data-ATS'
        )
        endpoint = res["endpointAddress"]
        config_dir = "/greengrass/config"
        if os.path.isdir(config_dir):
            shutil.rmtree(config_dir)  # remove dir and all contains
            self.__logger.info("Removed old config")

        os.mkdir(config_dir)
        default_config = {
            "coreThing": {
                "caPath": "root.ca.pem",
                "certPath": "%s.cert.pem" % self.__hash,
                "keyPath": "%s.private.key" % self.__hash,
                "thingArn": self.__thing_arn,
                "iotHost": endpoint,
                "ggHost": "greengrass-ats.iot.%s.amazonaws.com" % AWS.region(),
                "keepAlive": self.__KEEP_ALIVE,
                "ggDaemonPort": 8000,
                "systemComponentAuthTimeout": 5000
            },
            "runtime": {
                "maxWorkItemCount": 1024,
                "cgroup": {
                    "useSystemd": "yes"
                }
            },
            "managedRespawn": "false",
            "crypto": {
                "principals": {
                    "SecretsManager": {
                        "privateKeyPath": "file:///greengrass/certs/%s.private.key" % self.__hash
                    },
                    "IoTCertificate": {
                        "privateKeyPath": "file:///greengrass/certs/%s.private.key" % self.__hash,
                        "certificatePath": "file:///greengrass/certs/%s.cert.pem" % self.__hash
                    }
                },
                "caPath": "file:///greengrass/certs/root.ca.pem"
            }
        }

        f = open(config_dir + "/config.json", "w")
        f.write(json.dumps(default_config))
        f.close()

        self.__logger.info("Config saved")
        self.__logger.info(json.dumps(default_config))


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, *args, **kwargs):
        jdict = OrderedDict()

        jdict['core-name'] = self.core_name
        jdict['group-name'] = self.group_name

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def core_name(self):
        return self.__core_name


    @property
    def group_name(self):
        return self.__group_name


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSIdentity:{core_name:%s, group_name:%s}" % (self.core_name, self.group_name)




"""
Created on 8 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class PressureDatum(JSONable):
    """
    classdocs
    """

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        actual_press = jdict.get('pA')
        sl_press = jdict.get('p0')
        temp = jdict.get('tmp')

        return cls(actual_press, sl_press, temp)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def _sl_press(cls, actual_press, temp, altitude):
        if temp is None or altitude is None:
            return None

        return (actual_press * 9.80665 * altitude) / (287 * (273 + temp + (altitude / 400))) + actual_press


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, actual_press, sl_press, temp):
        """
        Constructor
        """
        self.__actual_press = Datum.float(actual_press, 1)          # kPa
        self.__sl_press = Datum.float(sl_press, 1)                  # kPa

        self.__temp = Datum.float(temp, 1)                          # °C


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['pA'] = self.actual_press

        if self.sl_press is not None:
            jdict['p0'] = self.sl_press

        if self.temp is not None:
            jdict['tmp'] = self.temp

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def actual_press(self):
        return self.__actual_press


    @property
    def sl_press(self):
        return self.__sl_press


    @property
    def temp(self):
        return self.__temp



"""
Created on 17 May 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"inhibit-publishing": false, "report-file": "/tmp/southcoastscience/mqtt_queue_report.json", "debug": true}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class MQTTConf(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "mqtt_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return MQTTConf(False, None, False) if skeleton else None

        inhibit_publishing = jdict.get('inhibit-publishing', False)
        report_file = jdict.get('report-file', None)
        debug = jdict.get('debug', False)

        return MQTTConf(inhibit_publishing, report_file, debug)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, inhibit_publishing, report_file, debug):
        """
        Constructor
        """
        super().__init__()

        self.__inhibit_publishing = bool(inhibit_publishing)                # do not attempt to publish
        self.__report_file = report_file                                    # tmp file to store current queue length
        self.__debug = bool(debug)                                          # DEBUG log level


    def __eq__(self, other):
        try:
            return self.inhibit_publishing == other.inhibit_publishing and self.report_file == other.report_file and \
                   self.debug == other.debug

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def inhibit_publishing(self):
        return self.__inhibit_publishing


    @property
    def report_file(self):
        return self.__report_file


    @property
    def debug(self):
        return self.__debug


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['inhibit-publishing'] = self.inhibit_publishing
        jdict['report-file'] = self.report_file
        jdict['debug'] = self.debug

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTConf:{inhibit_publishing:%s, report_file:%s, debug:%s}" %  \
               (self.inhibit_publishing, self.report_file, self.debug)



"""
Created on 16 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Resources:
https://ctlsys.com/support/how_to_compute_the_modbus_rtu_message_crc/

https://stackoverflow.com/questions/52391412/python-crc8-calculation
https://www.raspberrypi.org/forums/viewtopic.php?t=186576
"""


# --------------------------------------------------------------------------------------------------------------------

def modbus_crc(datagram):
    crc = 0xffff
    polynomial = 0xa001

    for datum in datagram:
        crc ^= datum

        for _ in range(8):
            if crc & 0x0001:
                crc >>= 1
                crc ^= polynomial
            else:
                crc >>= 1

    return crc


def crc8(datagram, crc=0xff, polynomial=0x31):
    for datum in datagram:
        for _ in range(8):
            if (datum ^ crc) & 0x80:
                crc <<= 1
                crc ^= polynomial
            else:
                crc <<= 1

            datum <<= 1

    return crc & 0xff



"""
Created on 24 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.gas.sensor_calib import SensorCalib


# --------------------------------------------------------------------------------------------------------------------

class A4Calib(SensorCalib):
    """
    classdocs
    """

    # used for Digital Single Interface...
    DEFAULT_WE_ELECTRONIC_ZERO_MV =     300
    DEFAULT_WE_SENSOR_ZERO_MV =           0
    DEFAULT_WE_TOTAL_ZERO_MV =          300

    DEFAULT_AE_ELECTRONIC_ZERO_MV =     300
    DEFAULT_AE_SENSOR_ZERO_MV =           0
    DEFAULT_AE_TOTAL_ZERO_MV =          300

    DEFAULT_PCB_GAIN =                 -0.7


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        sensor_type = jdict.get('sensor_type')

        we_elc_mv = jdict.get('we_electronic_zero_mv')
        we_cal_mv = jdict.get('we_sensor_zero_mv')
        we_tot_mv = jdict.get('we_total_zero_mv')

        ae_elc_mv = jdict.get('ae_electronic_zero_mv')
        ae_cal_mv = jdict.get('ae_sensor_zero_mv')
        ae_tot_mv = jdict.get('ae_total_zero_mv')

        we_sens_na = jdict.get('we_sensitivity_na_ppb')
        we_x_sens_na = jdict.get('we_cross_sensitivity_no2_na_ppb')

        pcb_gain = jdict.get('pcb_gain')

        we_sens_mv = jdict.get('we_sensitivity_mv_ppb')
        we_no2_x_sens_mv = jdict.get('we_cross_sensitivity_no2_mv_ppb')

        return cls(serial_number, sensor_type, we_elc_mv, we_cal_mv, we_tot_mv, ae_elc_mv, ae_cal_mv, ae_tot_mv,
                   we_sens_na, we_x_sens_na, pcb_gain, we_sens_mv, we_no2_x_sens_mv)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, sensor_type, we_elc_mv, we_cal_mv, we_tot_mv, ae_elc_mv, ae_cal_mv, ae_tot_mv,
                 we_sens_na, we_x_sens_na, pcb_gain, we_sens_mv, we_no2_x_sens_mv):
        """
        Constructor
        """
        SensorCalib.__init__(self, serial_number, sensor_type)

        self.__we_elc_mv = Datum.int(we_elc_mv)                     # WE electronic zero                    mV
        self.__we_cal_mv = Datum.int(we_cal_mv)                     # WE sensor zero at 23 °C               mV
        self.__we_tot_mv = Datum.int(we_tot_mv)                     # total WE zero                         mV

        self.__ae_elc_mv = Datum.int(ae_elc_mv)                     # Aux electronic zero                   mV
        self.__ae_cal_mv = Datum.int(ae_cal_mv)                     # Aux sensor zero at 23 °C              mV
        self.__ae_tot_mv = Datum.int(ae_tot_mv)                     # total Aux zero                        mV

        self.__we_sens_na = Datum.float(we_sens_na, 3)              # WE sensitivity                        nA
        self.__we_x_sens_na = Datum.float(we_x_sens_na, 3)          # WE cross-sensitivity                  nA

        self.__pcb_gain = Datum.float(pcb_gain, 3)                  # PCB gain                              mv / nA

        self.__we_sens_mv = Datum.float(we_sens_mv, 3)              # WE sensitivity                        mV / ppb
        self.__we_no2_x_sens_mv = Datum.float(we_no2_x_sens_mv, 3)  # WE cross-sensitivity                  mV / ppb


    def __eq__(self, other):
        try:
            return self.serial_number == other.serial_number and self.sensor_type == other.sensor_type and \
                   self.we_elc_mv == other.we_elc_mv and self.we_cal_mv == other.we_cal_mv and \
                   self.we_tot_mv == other.we_tot_mv and \
                   self.ae_elc_mv == other.ae_elc_mv and self.ae_cal_mv == other.ae_cal_mv and \
                   self.ae_tot_mv == other.ae_tot_mv and \
                   self.we_sens_na == other.we_sens_na and self.we_x_sens_na == other.we_x_sens_na and \
                   self.pcb_gain == other.pcb_gain and \
                   self.we_sens_mv == other.we_sens_mv and self.we_no2_x_sens_mv == other.we_no2_x_sens_mv

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def set_defaults(self):
        self.__we_elc_mv = self.DEFAULT_WE_ELECTRONIC_ZERO_MV
        self.__we_cal_mv = self.DEFAULT_WE_SENSOR_ZERO_MV
        self.__we_tot_mv = self.DEFAULT_WE_TOTAL_ZERO_MV

        self.__ae_elc_mv = self.DEFAULT_AE_ELECTRONIC_ZERO_MV
        self.__ae_cal_mv = self.DEFAULT_AE_SENSOR_ZERO_MV
        self.__ae_tot_mv = self.DEFAULT_AE_TOTAL_ZERO_MV

        self.__pcb_gain = self.DEFAULT_PCB_GAIN


    def set_sens_mv_from_sens_na(self):
        we_sens_mv = -0.7313 * self.__we_sens_na + -0.0006          # coefficients found from Alphasense calibrations
        self.__we_sens_mv = round(we_sens_mv, 3)


    def reports_no2_cross_sensitivity(self):
        return self.__we_no2_x_sens_mv is not None


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['sensor_type'] = self.sensor_type

        jdict['we_electronic_zero_mv'] = self.we_elc_mv
        jdict['we_sensor_zero_mv'] = self.we_cal_mv
        jdict['we_total_zero_mv'] = self.we_tot_mv

        jdict['ae_electronic_zero_mv'] = self.ae_elc_mv
        jdict['ae_sensor_zero_mv'] = self.ae_cal_mv
        jdict['ae_total_zero_mv'] = self.ae_tot_mv

        jdict['we_sensitivity_na_ppb'] = self.we_sens_na
        jdict['we_cross_sensitivity_no2_na_ppb'] = self.we_x_sens_na if self.we_x_sens_na else "n/a"

        jdict['pcb_gain'] = self.pcb_gain

        jdict['we_sensitivity_mv_ppb'] = self.we_sens_mv
        jdict['we_cross_sensitivity_no2_mv_ppb'] = self.we_no2_x_sens_mv if self.we_no2_x_sens_mv else "n/a"

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_elc_mv(self):
        return self.__we_elc_mv


    @property
    def we_cal_mv(self):
        return self.__we_cal_mv


    @property
    def we_tot_mv(self):
        return self.__we_tot_mv


    @property
    def ae_elc_mv(self):
        return self.__ae_elc_mv


    @property
    def ae_cal_mv(self):
        return self.__ae_cal_mv


    @property
    def ae_tot_mv(self):
        return self.__ae_tot_mv


    @property
    def we_sens_na(self):
        return self.__we_sens_na


    @property
    def we_x_sens_na(self):
        return self.__we_x_sens_na


    @property
    def pcb_gain(self):
        return self.__pcb_gain


    @property
    def we_sens_mv(self):
        return self.__we_sens_mv


    @we_sens_mv.setter
    def we_sens_mv(self, we_sens_mv):
        self.__we_sens_mv = we_sens_mv


    @property
    def we_no2_x_sens_mv(self):
        return self.__we_no2_x_sens_mv


    @we_no2_x_sens_mv.setter
    def we_no2_x_sens_mv(self, we_no2_x_sens_mv):
        self.__we_no2_x_sens_mv = we_no2_x_sens_mv


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Calib:{serial_number:%s, sensor_type:%s, we_elc_mv:%s, we_cal_mv:%s, we_tot_mv:%s, " \
               "ae_elc_mv:%s, ae_cal_mv:%s, ae_tot_mv:%s, we_sens_na:%s, we_x_sens_na:%s, pcb_gain:%s, " \
               "we_sens_mv:%s, we_no2_x_sens_mv:%s}" % \
                    (self.serial_number, self.sensor_type, self.we_elc_mv, self.we_cal_mv, self.we_tot_mv,
                     self.ae_elc_mv, self.ae_cal_mv, self.ae_tot_mv, self.we_sens_na, self.we_x_sens_na, self.pcb_gain,
                     self.we_sens_mv, self.we_no2_x_sens_mv)



"""
Created on 22 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense Application Note AAN 803-02
AAN 803-02 070916_DRAFT03.doc
"""

# import sys

from scs_core.gas.sensor import Sensor


# TODO: indicate with "alg" field to identify which equation is being used

# --------------------------------------------------------------------------------------------------------------------

class A4TempComp(object):
    """
    classdocs
    """

    __MIN_TEMP =        -30
    __MAX_TEMP =         50
    __INTERVAL =         10

    __COMP = None


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def init(cls):          # °C:                  -30   -20   -10    0    10    20    30    40    50
        cls.__COMP = {
            Sensor.CODE_CO:     A4TempComp(1, 'n_t', [1.0, 1.0, 1.0, 1.0, -0.2, -0.9, -1.5, -1.5, -1.5]),
            Sensor.CODE_H2S:    A4TempComp(1, 'n_t', [3.0, 3.0, 3.0, 1.0, -1.0, -2.0, -1.5, -1.0, -0.5]),
            Sensor.CODE_NO:     A4TempComp(3, 'kp_t', [0.7, 0.7, 0.7, 0.7, 0.8, 1.0, 1.2, 1.4, 1.6]),
            Sensor.CODE_NO2:    A4TempComp(1, 'n_t', [0.8, 0.8, 1.0, 1.2, 1.6, 1.8, 1.9, 2.5, 3.6]),
            Sensor.CODE_OX:     A4TempComp(3, 'kp_t', [0.1, 0.1, 0.2, 0.3, 0.7, 1.0, 1.7, 3.0, 4.0]),
            Sensor.CODE_SO2:    A4TempComp(1, 'kpp_t', [1.6, 1.6, 1.6, 1.6, 1.6, 1.6, 1.9, 3.0, 5.8]),

            Sensor.CODE_VOCe:   A4TempComp(1, 'n_t', [1.0, 1.0, 1.0, 1.0, -0.2, -0.9, -1.5, -1.5, -1.5]),

            Sensor.CODE_TEST_1: None,
            Sensor.CODE_TEST_2: None,
            Sensor.CODE_TEST_3: None,
            Sensor.CODE_TEST_4: None
        }

        #   Recommended, but causes div by zero error if calib.ae_cal_mv is zero
        #   Sensor.CODE_H2S:    A4TempComp(2, 'k_t', [-1.5, -1.5, -1.5, -0.5, 0.5, 1.0, 0.8, 0.5, 0.3]),


    @classmethod
    def find(cls, sensor_code):
        if sensor_code not in cls.__COMP:
            raise ValueError("A4TempComp.find: unrecognised sensor code: %s." % sensor_code)

        return cls.__COMP[sensor_code]


    @classmethod
    def in_range(cls, temp):
        if temp is None:
            return False

        return temp <= cls.__MAX_TEMP


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, algorithm, factor, values):
        """
        Constructor
        """
        length = (A4TempComp.__MAX_TEMP - A4TempComp.__MIN_TEMP) // A4TempComp.__INTERVAL + 1

        if len(values) != length:
            raise ValueError("A4TempComp: value count should be %d." % length)

        self.__algorithm = algorithm        # int
        self.__factor = factor              # string
        self.__values = values              # array of float


    # ----------------------------------------------------------------------------------------------------------------

    def correct(self, calib, temp, we_t, ae_t):
        """
        Compute weC from weT, aeT
        """
        if not A4TempComp.in_range(temp):
            return None

        if self.__algorithm == 1:
            return self.__eq1(temp, we_t, ae_t)

        if self.__algorithm == 2:
            return self.__eq2(temp, we_t, ae_t, calib.we_cal_mv, calib.ae_cal_mv)

        if self.__algorithm == 3:
            return self.__eq3(temp, we_t, ae_t, calib.we_cal_mv, calib.ae_cal_mv)

        if self.__algorithm == 4:
            return self.__eq4(temp, we_t, calib.we_cal_mv)

        raise ValueError("A4TempComp.conv: unrecognised algorithm: %d." % self.__algorithm)


    # ----------------------------------------------------------------------------------------------------------------

    def __eq1(self, temp, we_t, ae_t):
        n_t = self.cf_t(temp)

        we_c = we_t - n_t * ae_t

        # print("A4TempComp.__eq1: alg:%d, temp:%f we_t:%f n_t:%f we_c:%f " %
        #       (self.__algorithm, temp, we_t, n_t, we_c), file=sys.stderr)

        return we_c


    def __eq2(self, temp, we_t, ae_t, we_cal_mv, ae_cal_mv):
        k_t = self.cf_t(temp)

        we_c = we_t - k_t * (we_cal_mv / ae_cal_mv) * ae_t

        # print("A4TempComp.__eq2: alg:%d, temp:%f we_t:%f ae_t:%f we_cal_mv:%f ae_cal_mv:%f k_t:%f we_c:%f " %
        #       (self.__algorithm, temp, we_t, ae_t, we_cal_mv, ae_cal_mv, k_t, we_c), file=sys.stderr)

        return we_c


    def __eq3(self, temp, we_t, ae_t, we_cal_mv, ae_cal_mv):
        kp_t = self.cf_t(temp)

        we_c = we_t - kp_t * (we_cal_mv - ae_cal_mv) * ae_t

        # print("A4TempComp.__eq3: alg:%d, temp:%f we_t:%f ae_t:%f we_cal_mv:%f ae_cal_mv:%f kp_t:%f we_c:%f " %
        #       (self.__algorithm, temp, we_t, ae_t, we_cal_mv, ae_cal_mv, kp_t, we_c), file=sys.stderr)

        return we_c


    def __eq4(self, temp, we_t, we_cal_mv):
        kpp_t = self.cf_t(temp)

        we_c = we_t - we_cal_mv - kpp_t

        # print("A4TempComp.__eq4: alg:%d, temp:%f we_t:%f we_cal_mv:%f kpp_t:%f we_c:%f " %
        #       (self.__algorithm, temp, we_t, we_cal_mv, kpp_t, we_c), file=sys.stderr)

        return we_c


    # ----------------------------------------------------------------------------------------------------------------

    def cf_t(self, temp):
        """
        Compute the linear-interpolated temperature compensation factor.
        """
        # below MIN_TEMP...
        if temp < A4TempComp.__MIN_TEMP:
            return self.__values[0]

        index = int((temp - A4TempComp.__MIN_TEMP) // A4TempComp.__INTERVAL)        # index of start of interval

        # on boundary...
        if temp % A4TempComp.__INTERVAL == 0:
            return self.__values[index]

        # all others...
        y1 = self.__values[index]                                                   # y value at start of interval
        y2 = self.__values[index + 1]                                               # y value at end of interval

        delta_y = y2 - y1

        delta_x = float(temp % A4TempComp.__INTERVAL) / A4TempComp.__INTERVAL       # proportion of interval

        cf_t = y1 + (delta_y * delta_x)

        # print("A4TempComp.cf_t: alg:%d, temp:%f y1:%f y2:%f delta_y:%f delta_x:%f cf_t:%f " %
        #       (self.__algorithm, temp, y1, y2, delta_y, delta_x, cf_t), file=sys.stderr)

        return cf_t


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def algorithm(self):
        return self.__algorithm


    @property
    def factor(self):
        return self.__factor


    @property
    def values(self):
        return self.__values


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4TempComp:{algorithm:%d, factor:%s, values:%s}" % (self.algorithm, self.factor, self.values)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://xy1eszuu23.execute-api.us-west-2.amazonaws.com/staging/topicMessages?
topic=south-coast-science-dev/production-test/loc/1/gases&
startTime=2018-03-31T07:50:59.712Z&
endTime=2018-03-31T07:55:59.712Z

header:
CURLOPT_HTTPHEADER => array('Accept: application/json', 'Authorization: api-key ' . AWS_API_KEY),
"""

import json

from scs_core.data.json import JSONify

from scs_core.client.http_client import HTTPClient

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.http_status import HTTPStatus


# --------------------------------------------------------------------------------------------------------------------

class RESTClient(object):
    """
    classdocs
    """

    __HEADER_ACCEPT = "application/json"
    __HEADER_AUTHORIZATION = "api-key "

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, auth):
        """
        Constructor
        """
        self.__http_client = HTTPClient()
        self.__auth = auth


    # ----------------------------------------------------------------------------------------------------------------

    def connect(self):
        self.__http_client.connect(self.__auth.endpoint)


    def close(self):
        self.__http_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def get(self, path, params=None):
        # print("RESTClient.get: path: %s params: %s headers:%s" % (path, params, self.__headers), file=sys.stderr)

        try:
            response_jstr = self.__http_client.get(path, params, self.__headers)
            # print("response_jstr: %s" % response_jstr, file=sys.stderr)
            # print("-", file=sys.stderr)

        except HTTPException as exc:
            # print("exc: %s" % exc, file=sys.stderr)
            # print("-", file=sys.stderr)

            if exc.status == HTTPStatus.NOT_FOUND:
                return None

            else:
                raise exc

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def post(self, path, payload_jdict):
        payload_jstr = JSONify.dumps(payload_jdict)

        # print("RESTClient.post: path: %s payload: %s" % (path, payload_jstr))

        response_jstr = self.__http_client.post(path, payload_jstr, self.__headers)

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def put(self, path, payload_jdict):
        payload_jstr = JSONify.dumps(payload_jdict)

        # print("RESTClient.put: path: %s payload: %s" % (path, payload_jstr))

        response_jstr = self.__http_client.put(path, payload_jstr, self.__headers)

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def delete(self, path):
        response_jstr = self.__http_client.delete(path, self.__headers)

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def __headers(self):
        return {"Accept": RESTClient.__HEADER_ACCEPT,
                "Authorization": RESTClient.__HEADER_AUTHORIZATION + self.__auth.api_key}


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "RESTClient:{http_client:%s, auth:%s}" % (self.__http_client, self.__auth)



"""
Created on 4 May 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"serial_number": "27-000001", "type": "810-0023-02", "calibrated_on": "2016-11-01",
"sn1": {"serial_number": "212060308", "sensor_type": "NO2A43F"},
"sn2": {"serial_number": "132950202", "sensor_type": "CO A4"},
"sn3": {"serial_number": "134060009", "sensor_type": "SO2A4"},
"sn4": {"serial_number": "133910023", "sensor_type": "H2SA4"}}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable, PersistentJSONable
from scs_core.data.str import Str

from scs_core.gas.afe_calib import AFECalib
from scs_core.gas.sensor import Sensor


# --------------------------------------------------------------------------------------------------------------------

class AFEId(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def persistence_location(cls):
        return AFECalib.persistence_location()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        afe_type = jdict.get('type')
        calibrated_on = Datum.date(jdict.get('calibrated_on'))

        sensor_ids = []

        for key in sorted(jdict.keys()):
            if key[:2] == "sn":
                if jdict[key] is None:
                    sensor_ids.append(None)
                    continue

                sensor_ids.append(SensorId.construct_from_jdict(jdict[key]))

        return cls(serial_number, afe_type, calibrated_on, sensor_ids)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, afe_type, calibrated_on, sensor_ids):
        """
        Constructor
        """
        super().__init__()

        self.__serial_number = serial_number            # string
        self.__afe_type = afe_type                      # string
        self.__calibrated_on = calibrated_on            # date

        self.__sensor_ids = sensor_ids                  # array of SensorId


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for i in range(len(self)):
                if self.sensor_id(i) != other.sensor_id(i):
                    return False

            return self.serial_number == other.serial_number and self.afe_type == other.afe_type and \
                self.calibrated_on == other.calibrated_on

        except (AttributeError, TypeError):
            return False


    def __len__(self):
        return len(self.__sensor_ids)


    # ----------------------------------------------------------------------------------------------------------------

    def sensor_index(self, gas_name):
        for i in range(len(self.__sensor_ids)):
            sensor_id = self.__sensor_ids[i]

            if sensor_id is None:
                continue

            sensor = Sensor.find(sensor_id.serial_number)

            if sensor is None:
                raise ValueError(sensor_id.serial_number)

            if sensor.gas_name == gas_name:
                return i

        return None


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['type'] = self.afe_type
        jdict['calibrated_on'] = self.calibrated_on.isoformat() if self.calibrated_on else None

        for i in range(len(self.__sensor_ids)):
            jdict['sn' + str(i + 1)] = self.__sensor_ids[i]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def serial_number(self):
        return self.__serial_number


    @property
    def afe_type(self):
        return self.__afe_type


    @property
    def calibrated_on(self):
        return self.__calibrated_on


    def sensor_id(self, i):
        return self.__sensor_ids[i]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AFEId:{serial_number:%s, afe_type:%s, calibrated_on:%s, sensor_ids:%s}" %  \
                     (self.serial_number, self.afe_type, self.calibrated_on, Str.collection(self.__sensor_ids))


# --------------------------------------------------------------------------------------------------------------------

class SensorId(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        sensor_type = jdict.get('sensor_type')

        return cls(serial_number, sensor_type)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, sensor_type):
        """
        Constructor
        """
        self.__serial_number = serial_number                    # int
        self.__sensor_type = sensor_type                        # string


    def __eq__(self, other):
        try:
            return self.serial_number == other.serial_number and self.sensor_type == other.sensor_type

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['sensor_type'] = self.sensor_type

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def serial_number(self):
        return self.__serial_number


    @property
    def sensor_type(self):
        return self.__sensor_type


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SensorId:{serial_number:%s, sensor_type:%s}" %  (self.serial_number, self.sensor_type)



"""
Created on 2 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

WARNING: Path methods require that the __HOST field must be instantiated.

example document:
{"endpoint": "endpoint.iot.us-west-2.amazonaws.com", "client-id": "bruno", "cert-id": "9f08402232"}
"""

import os

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class ClientAuth(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "aws_client_auth.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    @classmethod
    def load(cls, manager, encryption_key=None, skeleton=False):
        auth = super().load(manager, encryption_key=encryption_key, skeleton=skeleton)

        if auth is None:
            return None

        auth.manager = manager

        return auth


    # ----------------------------------------------------------------------------------------------------------------

    __CERT_DIR = "certs"                                # hard-coded path

    __ROOT_CA = "root-CA.crt"

    __PUBLIC_KEY_SUFFIX = "-public.pem.key"
    __PRIVATE_KEY_SUFFIX = "-private.pem.key"
    __CERT_SUFFIX = "-certificate.pem.crt"


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        endpoint = jdict.get('endpoint')
        client_id = jdict.get('client-id')
        cert_id = jdict.get('cert-id')

        return ClientAuth(endpoint, client_id, cert_id)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, endpoint, client_id, cert_id, manager=None):
        """
        Constructor
        """
        super().__init__()

        self.__endpoint = endpoint                  # String
        self.__client_id = client_id                # String
        self.__cert_id = cert_id                    # String

        self.__manager = manager                    # FilesystemPersistenceManager


    def __eq__(self, other):                # ignore manager
        try:
            return self.endpoint == other.endpoint and self.client_id == other.client_id and \
                   self.cert_id == other.cert_id

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['endpoint'] = self.endpoint
        jdict['client-id'] = self.client_id
        jdict['cert-id'] = self.cert_id

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def installed_ok(self):
        if not os.access(self.root_ca_file_path, os.R_OK):
            return False

        if not os.access(self.certificate_path, os.R_OK):
            return False

        if not os.access(self.public_key_path, os.R_OK):
            return False

        if not os.access(self.private_key_path, os.R_OK):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def root_ca_file_path(self):
        return os.path.join(self.__cert_path(), self.__ROOT_CA)


    @property
    def certificate_path(self):
        return os.path.join(self.__cert_path(), self.cert_id + self.__CERT_SUFFIX)


    @property
    def public_key_path(self):
        return os.path.join(self.__cert_path(), self.cert_id + self.__PUBLIC_KEY_SUFFIX)


    @property
    def private_key_path(self):
        return os.path.join(self.__cert_path(), self.cert_id + self.__PRIVATE_KEY_SUFFIX)


    # ----------------------------------------------------------------------------------------------------------------

    def __cert_path(self):
        return os.path.join(self.manager.scs_path(), self.aws_dir(), self.__CERT_DIR)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def endpoint(self):
        return self.__endpoint


    @property
    def client_id(self):
        return self.__client_id


    @property
    def cert_id(self):
        return self.__cert_id


    @property
    def manager(self):
        return self.__manager


    @manager.setter
    def manager(self, manager):
        self.__manager = manager


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ClientAuth:{endpoint:%s, client_id:%s, cert_id:%s, manager:%s}" % \
               (self.endpoint, self.client_id, self.cert_id, self.manager)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
"location": {
    "lat": 50.819456,
    "lon": -0.128336,
    "postcode": "bn2 1af"
}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Location(JSONable):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        lat = jdict.get('lat')
        lng = jdict.get('lon')

        elevation = jdict.get('elevation')

        zipcode = jdict.get('zip')
        postcode = jdict.get('postcode')

        return Location(lat, lng, elevation, zipcode, postcode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, lat, lng, elevation, zipcode, postcode):
        """
        Constructor
        """
        self.__lat = lat                    # float
        self.__lng = lng                    # float

        self.__elevation = elevation        # int

        self.__zip = zipcode                # string
        self.__postcode = postcode          # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['lat'] = self.lat
        jdict['lon'] = self.lng

        if self.elevation is not None:
            jdict['elevation'] = self.elevation

        if self.zip:
            jdict['zip'] = self.zip

        jdict['postcode'] = self.postcode

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def lat(self):
        return self.__lat


    @property
    def lng(self):
        return self.__lng


    @property
    def elevation(self):
        return self.__elevation


    @property
    def zip(self):
        return self.__zip


    @property
    def postcode(self):
        return None if self.__postcode is None else self.__postcode.upper()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Location:{lat:%s, lng:%s, elevation:%s, zip:%s, postcode:%s}" % \
                        (self.lat, self.lng, self.elevation, self.zip, self.postcode)



"""
Created on 17 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"rec": "2021-10-11T10:52:06Z", "tag": "scs-be2-3", "ver": 2.0, "val": {"hmd": 52.8, "tmp": 21.9,
"bar": {"pA": 103.5, "p0": 104.7}}}
"""

from collections import OrderedDict

from scs_core.climate.pressure_datum import PressureDatum
from scs_core.climate.sht_datum import SHTDatum

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.str import Str

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ClimateSample(Sample):
    """
    classdocs
    """

    VERSION = 1.0

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # Sample...
        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        version = jdict.get('ver', cls.DEFAULT_VERSION)

        val = jdict.get('val')
        exegeses = jdict.get('exg')

        # ClimateSample...
        sht_datum = SHTDatum(val.get('hmd'), val.get('tmp'))

        node = val.get('bar')
        barometer_datum = None if node is None else PressureDatum(node.get('pA'), node.get('p0'), None)

        return cls(tag, rec, sht_datum, barometer_datum, version=version, exegeses=exegeses)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, sht_datum, barometer_datum, version=None, exegeses=None):
        """
        Constructor
        """
        if version is None:
            version = self.VERSION

        super().__init__(tag, rec, version, exegeses=exegeses)

        self.__sht_datum = sht_datum                                # SHT31Datum
        self.__barometer_datum = barometer_datum                    # PressureDatum


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def has_invalid_value(cls):
        # TODO: implement has_invalid_value
        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        jdict = OrderedDict()

        jdict['hmd'] = self.sht_datum.humid
        jdict['tmp'] = self.sht_datum.temp

        jdict['bar'] = self.barometer_datum

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sht_datum(self):
        return self.__sht_datum


    @property
    def barometer_datum(self):
        return self.__barometer_datum


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        exegeses = Str.collection(self.exegeses)

        return "ClimateSample:{tag:%s, rec:%s, exegeses:%s, sht_datum:%s, barometer_datum:%s}" % \
            (self.tag, self.rec, exegeses, self.sht_datum, self.barometer_datum)



"""
Created on 16 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/2520893/how-to-flush-the-input-stream-in-python
"""

import json
import os
import sys
import termios

from collections import OrderedDict
from getpass import getpass

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class AccessKey(PersistentJSONable):
    """
    classdocs
    """

    __ID_NAME = 'AWS_ACCESS_KEY_ID'
    __SECRET_NAME = 'AWS_SECRET_ACCESS_KEY'

    __FILENAME = "access_key.json"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def from_stdin(cls):
        line = sys.stdin.readline()
        jdict = json.loads(line)

        return cls.construct_from_jdict(jdict)


    @classmethod
    def from_environment(cls):
        if cls.__ID_NAME not in os.environ or cls.__SECRET_NAME not in os.environ:
            return None

        id = os.environ[cls.__ID_NAME]
        secret = os.environ[cls.__SECRET_NAME]

        return cls(id, secret)


    @classmethod
    def from_user(cls):
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)           # flush stdin
        except termios.error:
            pass

        print("Enter AWS Access Key ID: ", end="", file=sys.stderr)
        id = input().strip()

        print("Enter AWS Secret Access Key: ", end="", file=sys.stderr)
        secret = input().strip()

        return cls(id, secret)


    @staticmethod
    def password_from_user():
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)           # flush stdin
        except termios.error:
            pass

        return getpass().strip()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        id = jdict.get('key-id')
        secret = jdict.get('secret-key')

        return cls(id, secret)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, secret):
        """
        Constructor
        """
        super().__init__()

        self.__id = id                              # string
        self.__secret = secret                      # string


    # ----------------------------------------------------------------------------------------------------------------

    def ok(self):
        return self.id and self.secret


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['key-id'] = self.id
        jdict['secret-key'] = self.secret

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def secret(self):
        return self.__secret


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AccessKey:{id:%s, secret:%s}" %  (self.id, self.secret)



"""
Created on 7 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/ScheduledEvents.html

document example:
{"interval": 1, "units": "D"}
"""

import pytz

from abc import abstractmethod
from collections import OrderedDict
from datetime import datetime

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class RecurringPeriod(JSONable):
    """
    classdocs
    """
    _UTC = pytz.timezone('Etc/UTC')

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def valid_intervals(cls):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        interval = jdict.get('interval')
        units = jdict.get('units')

        return cls.construct(interval, units)


    @classmethod
    def construct(cls, interval, units):
        if units == 'D':
            return RecurringDay(interval)

        if units == 'H':
            return RecurringHours(interval)

        if units == 'M':
            return RecurringMinutes(interval)

        raise ValueError(units)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, interval):
        """
        Constructor
        """
        self.__interval = int(interval)                     # int


    @abstractmethod
    def __lt__(self, other):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def is_valid(self):
        pass


    @abstractmethod
    def checkpoint(self):
        pass


    @abstractmethod
    def cron(self, minutes_offset):
        pass


    @abstractmethod
    def aws_cron(self, minutes_offset):
        pass


    @abstractmethod
    def timedelta(self):
        pass


    @abstractmethod
    def end_datetime(self, origin: LocalizedDatetime):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['interval'] = self.interval
        jdict['units'] = self.units

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def interval(self):
        return self.__interval


    @property
    @abstractmethod
    def units(self):
        pass


    @property
    @abstractmethod
    def repr_units(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def __repr__(self):
        return ' '.join((str(self.interval), self.repr_units))


    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{interval:%s}" %  self.interval


# --------------------------------------------------------------------------------------------------------------------

class RecurringDay(RecurringPeriod):
    """
    classdocs
    """

    @classmethod
    def valid_intervals(cls):
        return (1, )


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, interval):
        """
        Constructor
        """
        super().__init__(interval)


    def __lt__(self, other):
        if isinstance(other, RecurringHours) or isinstance(other, RecurringMinutes):
            return False

        return self.interval < other.interval


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        return self.interval == 1


    def checkpoint(self):
        return '00:00:00'


    def cron(self, minutes_offset):
        return '%d 0 * * *' % minutes_offset


    def aws_cron(self, minutes_offset):
        return 'cron(%d 0 * * ? *)' % minutes_offset


    def timedelta(self):
        return Timedelta(days=self.interval)


    def end_datetime(self, origin: LocalizedDatetime):
        dt = origin.utc().datetime

        return LocalizedDatetime(datetime(dt.year, month=dt.month, day=dt.day, tzinfo=self._UTC))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def units(self):
        return 'D'


    @property
    def repr_units(self):
        return 'day' if self.interval == 1 else 'days'


# --------------------------------------------------------------------------------------------------------------------

class RecurringHours(RecurringPeriod):
    """
    classdocs
    """

    __DIVISORS = (1, 2, 3, 4, 6, 8, 12)

    @classmethod
    def valid_intervals(cls):
        return cls.__DIVISORS


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, interval):
        """
        Constructor
        """
        super().__init__(interval)


    def __lt__(self, other):
        if isinstance(other, RecurringDay):
            return True

        if isinstance(other, RecurringMinutes):
            return False

        return self.interval < other.interval


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        return self.interval in self.__DIVISORS


    def checkpoint(self):
        return '/%d:00:00' % self.interval


    def cron(self, minutes_offset):
        return '%d */%d * * *' % (minutes_offset, self.interval)


    def aws_cron(self, minutes_offset):
        return 'cron(%d 0/%d * * ? *)' % (minutes_offset, self.interval)


    def timedelta(self):
        return Timedelta(hours=self.interval)


    def end_datetime(self, origin: LocalizedDatetime):
        dt = origin.utc().datetime
        hour = (dt.hour // self.interval) * self.interval

        return LocalizedDatetime(datetime(dt.year, month=dt.month, day=dt.day, hour=hour, tzinfo=self._UTC))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def units(self):
        return 'H'


    @property
    def repr_units(self):
        return 'hour' if self.interval == 1 else 'hours'


# --------------------------------------------------------------------------------------------------------------------

class RecurringMinutes(RecurringPeriod):
    """
    classdocs
    """

    __DIVISORS = (1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30)

    @classmethod
    def valid_intervals(cls):
        return cls.__DIVISORS


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, interval):
        """
        Constructor
        """
        super().__init__(interval)


    def __lt__(self, other):
        if isinstance(other, RecurringDay) or isinstance(other, RecurringHours):
            return True

        return self.interval < other.interval


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        return self.interval in self.__DIVISORS


    def checkpoint(self):
        return '**:/%d:00' % self.interval


    def cron(self, minutes_offset):
        minutes = [str(minute + minutes_offset) for minute in range(0, 60, self.interval)]

        return '%s * * * *' % ','.join(minutes)


    def aws_cron(self, minutes_offset):
        minutes = '/'.join((str(minutes_offset), str(self.interval)))

        return 'cron(%s * * * ? *)' % minutes


    def timedelta(self):
        return Timedelta(minutes=self.interval)


    def end_datetime(self, origin: LocalizedDatetime):
        dt = origin.utc().datetime
        minute = (dt.minute // self.interval) * self.interval

        return LocalizedDatetime(datetime(dt.year, month=dt.month, day=dt.day, hour=dt.hour, minute=minute,
                                          tzinfo=self._UTC))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def units(self):
        return 'M'


    @property
    def repr_units(self):
        return 'minute' if self.interval == 1 else 'minutes'
